# adapted from https://github.com/pyta-uoft/pyta/blob/4c858623549e24a49fea7aef9c8ec7c20c836bd6/python_ta/cfg/visitor.py

from typing import Any, Dict, List, Optional, Tuple, Union

from astroid import nodes

from edulint.linting.analyses.cfg.graph import CFGBlock, ControlFlowGraph


class CFGVisitor:
    """An astroid visitor that creates a control flow graph for a given Python module.

    Initialize with an options dict to configure the visitor.

    Supported Options:
      - "separate-condition-blocks": bool
            This option specifies whether the test condition of an if statement gets merged with any
            preceding statements or placed in a new block. By default, it will merge them.
      - "functions": list
            This option specifies whether to restrict the creation of cfgs to just top-level
            function definitions or methods provided in this list. By default, it will create the
            cfg for the file as it normally would.

    Private Attributes:
    _control_boundaries: A stack of the boundaries the visitor is currently in.
        The top of the stack corresponds to the end of the list.
        (compound statement [while], {'Break'/'Continue': CFGBlock to link to})
    """

    cfgs: Dict[Union[nodes.FunctionDef, nodes.Module], ControlFlowGraph]
    options: Dict[str, Any]
    # cfg_count is used as an "auto-increment" to ensure cfg ids are unique.
    cfg_count: int
    _current_cfg: Optional[ControlFlowGraph]
    _current_block: Optional[CFGBlock]
    _control_boundaries: List[Tuple[nodes.NodeNG, Dict[str, CFGBlock]]]

    def __init__(self, options: Optional[Dict[str, Any]] = None) -> None:
        super().__init__()
        self.cfgs = {}
        self.options = {
            "separate-condition-blocks": False,
            "functions": [],
        }
        if options is not None:
            self.options.update(options)
        self.cfg_count = 0
        self._current_cfg = None
        self._current_block = None
        self._control_boundaries = []

    def __getattr__(self, attr: str):
        if attr.startswith("visit_"):
            return self.visit_generic
        else:
            raise AttributeError(f"'CFGVisitor' object has not attribute '{attr}'")

    def visit_generic(self, node: nodes.NodeNG) -> None:
        """By default, add the expression to the end of the current block."""
        if self._current_block is not None:
            self._current_block.add_statement(node)

    def visit_module(self, module: nodes.Module) -> None:
        # Modules shouldn't be considered if user specifies functions to render
        functions_to_render = self.options.get("functions", [])
        if functions_to_render:
            for child in module.body:
                # Check any classes or function definitions for the target function/method
                if isinstance(child, nodes.FunctionDef) or isinstance(child, nodes.ClassDef):
                    child.accept(self)
            return

        self.cfgs[module] = ControlFlowGraph(self.cfg_count)
        self.cfg_count += 1
        self._current_cfg = self.cfgs[module]
        self._current_block = self._current_cfg.start

        for child in module.body:
            child.accept(self)

        self._current_cfg.start.add_CF_statement(module)

        self._current_cfg.link_or_merge(self._current_block, self._current_cfg.end)
        self._current_cfg.update_block_reachability()

    def visit_classdef(self, node: nodes.ClassDef) -> None:
        functions_to_render = self.options.get("functions", [])

        if self._current_block is not None:
            self._current_block.add_statement(node)

        previous_cfg = self._current_cfg
        previous_block = self._current_block

        self.cfgs[node] = ControlFlowGraph(self.cfg_count)
        self.cfg_count += 1
        self._current_cfg = self.cfgs[node]

        self._current_block = self._current_cfg.create_block()

        for child in node.body:
            if functions_to_render:
                if isinstance(child, nodes.FunctionDef):
                    child.accept(self)
            else:
                child.accept(self)

        self._current_cfg.link_or_merge(self._current_block, self._current_cfg.end)
        self._current_cfg.update_block_reachability()

        self._current_block = previous_block
        self._current_cfg = previous_cfg

    def visit_asyncfunctiondef(self, func: nodes.AsyncFunctionDef) -> None:
        self.visit_functiondef(func)

    def visit_functiondef(self, func: nodes.FunctionDef) -> None:
        # If user specifies to only render functions, check if the function/method name is listed
        functions_to_render = self.options.get("functions", [])
        scope_parent = func.scope().parent

        if functions_to_render:
            if (
                isinstance(scope_parent, nodes.ClassDef)
                and (scope_parent.name + "." + func.name) not in functions_to_render
            ) or (isinstance(scope_parent, nodes.Module) and func.name not in functions_to_render):
                return

        if self._current_block is not None:
            self._current_block.add_statement(func)

        previous_cfg = self._current_cfg
        previous_block = self._current_block

        self.cfgs[func] = ControlFlowGraph(self.cfg_count)
        self.cfg_count += 1
        self._current_cfg = self.cfgs[func]

        self._control_boundaries.append(
            (
                func,
                {
                    nodes.Return.__name__: self._current_cfg.end,
                    nodes.Raise.__name__: self._current_cfg.end,
                },
            )
        )

        self._current_cfg.start.add_statement(func.args)

        self._current_block = self._current_cfg.create_block(self._current_cfg.start)

        for child in func.body:
            child.accept(self)

        self._control_boundaries.pop()

        self._current_cfg.link_or_merge(self._current_block, self._current_cfg.end)
        self._current_cfg.update_block_reachability()

        self._current_block = previous_block
        self._current_cfg = previous_cfg

    def visit_if(self, node: nodes.If) -> None:
        # When only creating cfgs for functions, _current_cfg will only be None outside of functions
        if self._current_cfg is None:
            return

        separate_conditions = self.options.get("separate-condition-blocks", False)
        if separate_conditions:
            # Create a block for the test condition
            test_block = self._current_cfg.create_block(self._current_block)
            test_block.add_statement(node.test)
            self._current_block = test_block
        else:
            # If the options doesn't specify to separate the test condition blocks, just add it to
            # the current block.
            self._current_block.add_statement(node.test)
        self._current_block.add_CF_statement(node)
        old_curr = self._current_block

        # Handle "then" branch and label it.
        then_block = self._current_cfg.create_block(old_curr, edge_label="True")
        self._current_block = then_block
        for child in node.body:
            child.accept(self)
        end_if = self._current_block

        # Handle "else" branch.
        if node.orelse == []:
            end_else = old_curr
        else:
            # Label the edge to the else block.
            else_block = self._current_cfg.create_block(old_curr, edge_label="False")
            self._current_block = else_block
            for child in node.orelse:
                child.accept(self)
            end_else = self._current_block

        after_if_block = self._current_cfg.create_block()
        self._current_cfg.link_or_merge(end_if, after_if_block)
        # Label the edge if there was no "else" branch
        if node.orelse == []:
            self._current_cfg.link_or_merge(end_else, after_if_block, edge_label="False")
        else:
            self._current_cfg.link_or_merge(end_else, after_if_block)

        self._current_block = after_if_block

    def visit_while(self, node: nodes.While) -> None:
        # When only creating cfgs for functions, _current_cfg will only be None outside of functions
        if self._current_cfg is None:
            return

        old_curr = self._current_block

        # Handle "test" block
        test_block = self._current_cfg.create_block()
        test_block.add_statement(node.test)
        test_block.add_CF_statement(node)
        self._current_cfg.link_or_merge(old_curr, test_block)

        after_while_block = self._current_cfg.create_block()

        # step into while
        self._control_boundaries.append(
            (
                node,
                {nodes.Break.__name__: after_while_block, nodes.Continue.__name__: test_block},
            )
        )

        # Handle "body" branch
        body_block = self._current_cfg.create_block(test_block, edge_label="True")
        self._current_block = body_block
        for child in node.body:
            child.accept(self)
        end_body = self._current_block
        self._current_cfg.link_or_merge(end_body, test_block)

        # step out of while
        self._control_boundaries.pop()

        # Handle "else" branch
        else_block = self._current_cfg.create_block(test_block, edge_label="False")
        self._current_block = else_block
        for child in node.orelse:
            child.accept(self)
        end_else = self._current_block

        self._current_cfg.link_or_merge(end_else, after_while_block)
        self._current_block = after_while_block

    def visit_for(self, node: nodes.For) -> None:
        # When only creating cfgs for functions, _current_cfg will only be None outside of functions
        if self._current_cfg is None:
            return

        old_curr = self._current_block
        old_curr.add_statement(node.iter)
        old_curr.add_CF_statement(node)

        # Handle "test" block
        test_block = self._current_cfg.create_block()
        test_block.add_statement(node.target)
        self._current_cfg.link_or_merge(old_curr, test_block)

        after_for_block = self._current_cfg.create_block()

        # step into for
        self._control_boundaries.append(
            (node, {nodes.Break.__name__: after_for_block, nodes.Continue.__name__: test_block})
        )

        # Handle "body" branch
        body_block = self._current_cfg.create_block(test_block)
        self._current_block = body_block
        for child in node.body:
            child.accept(self)
        end_body = self._current_block
        self._current_cfg.link_or_merge(end_body, test_block)

        # step out of for
        self._control_boundaries.pop()

        # Handle "else" branch
        else_block = self._current_cfg.create_block(test_block)
        self._current_block = else_block
        for child in node.orelse:
            child.accept(self)
        end_else = self._current_block

        self._current_cfg.link_or_merge(end_else, after_for_block)
        self._current_block = after_for_block

    def visit_break(self, node: nodes.Break) -> None:
        self._visit_jump(node)

    def visit_continue(self, node: nodes.Continue) -> None:
        self._visit_jump(node)

    def visit_return(self, node: nodes.Return) -> None:
        self._visit_jump(node)

    def visit_raise(self, node: nodes.Raise) -> None:
        self._visit_jump(node)

    def _visit_jump(
        self, node: Union[nodes.Break, nodes.Continue, nodes.Return, nodes.Raise]
    ) -> None:
        # When only creating cfgs for functions, _current_cfg will only be None outside of functions
        if self._current_cfg is None:
            return

        old_curr = self._current_block
        old_curr.add_statement(node)
        for boundary, exits in reversed(self._control_boundaries):
            if isinstance(node, nodes.Raise):
                exc_name = _get_raise_exc(node)

                if exc_name in exits:
                    self._current_cfg.link(old_curr, exits[exc_name])
                    break

            if type(node).__name__ in exits:
                self._current_cfg.link(old_curr, exits[type(node).__name__])
                break
        unreachable_block = self._current_cfg.create_block()
        self._current_block = unreachable_block

    def visit_try(self, node: nodes.Try) -> None:
        # When only creating cfgs for functions, _current_cfg will only be None outside of functions
        if self._current_cfg is None:
            return

        if self._current_block.locs != []:
            self._current_block = self._current_cfg.create_block(self._current_block)
        node_block = self._current_block

        # Construct the exception handlers first
        # Initialize a temporary block to later merge with end_body
        self._current_block = self._current_cfg.create_block()
        temp = self._current_block
        finally_block = self._current_cfg.create_block()
        end_block = self._current_cfg.create_block()
        # Case where Raise is not handled in the try block
        self._control_boundaries.append((node, {nodes.Raise.__name__: end_block}))
        cbs_added = 1

        after_body = []
        # Construct blocks in reverse to give precedence to the first block in overlapping except
        # branches
        for handler in reversed(node.handlers):
            h = self._current_cfg.create_block()
            self._current_block = h

            exceptions = _extract_exceptions(handler)
            # Edge case: catch-all except clause (i.e. except: ...)
            if exceptions == []:
                self._control_boundaries.append((node, {nodes.Raise.__name__: h}))
                cbs_added += 1

            # General case: specific except clause
            for exception in exceptions:
                self._control_boundaries.append((node, {f"{nodes.Raise.__name__} {exception}": h}))
                cbs_added += 1

            if handler.name is not None:  # The name assigned to the caught exception.
                handler.name.accept(self)
            for child in handler.body:
                child.accept(self)
            end_handler = self._current_block

            if len(h.locs) == 0:
                # h was overtaken by link_or_merge
                assert len(handler.body) > 0
                handler.body[0].cfg_loc.block.add_CF_statement(handler)
            else:
                h.add_CF_statement(handler)

            self._current_cfg.link_or_merge(end_handler, finally_block)
            after_body.append(h)

        if node.orelse == []:
            after_body.append(finally_block)
        else:
            self._current_block = self._current_cfg.create_block()
            after_body.append(self._current_block)
            for child in node.orelse:
                child.accept(self)
            self._current_cfg.link_or_merge(self._current_block, finally_block)

        # Construct the try body so reset current block to this node's block
        self._current_block = node_block

        for child in node.body:
            child.accept(self)
        end_body = self._current_block

        assert len(node.body) > 0
        # do not use node_block, because it may be overtaken by link_or_merge
        node.body[0].cfg_loc.block.add_CF_statement(node)

        # Remove each control boundary that we added in this method
        for _ in range(cbs_added):
            self._control_boundaries.pop()

        self._current_cfg.link_or_merge(temp, end_body)
        self._current_cfg.multiple_link_or_merge(end_body, after_body)
        self._current_cfg.link_or_merge(finally_block, end_block)
        self._current_block = end_block

    def visit_with(self, node: nodes.With) -> None:
        # When only creating cfgs for functions, _current_cfg will only be None outside of functions
        if self._current_cfg is None:
            return

        self._current_block.add_CF_statement(node)
        for context_node, name in node.items:
            self._current_block.add_statement(context_node)
            if name is not None:
                self._current_block.add_statement(name)

        for child in node.body:
            child.accept(self)


def _extract_exceptions(node: nodes.ExceptHandler) -> List[str]:
    """A helper method that returns a list of all the exceptions handled by this except block as a
    list of strings.
    """
    exceptions = node.type
    exceptions_so_far = []
    # ExceptHandler.type will either be Tuple, NodeNG, or None.
    if exceptions is None:
        return exceptions_so_far

    # Get all the Name nodes for all exceptions this except block is handling
    for exception in exceptions.nodes_of_class(nodes.Name):
        exceptions_so_far.append(exception.name)

    return exceptions_so_far


def _get_raise_exc(node: nodes.Raise) -> str:
    """A helper method that returns a string formatted for the control boundary representing the
    exception that this Raise node throws.

    Preconditions:
        - the raise statement is of the form 'raise' or 'raise <exception_class>'
    """
    exceptions = node.nodes_of_class(nodes.Name)

    # Return the formatted name of the exception or the just 'Raise' otherwise
    try:
        return f"{nodes.Raise.__name__} {next(exceptions).name}"
    except StopIteration:
        return nodes.Raise.__name__



import pytest
from edulint.edulint import main
from os.path import join
import sys
import json

def run(monkeypatch, capsys, argv) -> str:
    monkeypatch.setattr("sys.argv", ["script", "check"] + argv)
    main()
    captured = capsys.readouterr()

    print(captured.err, file=sys.stderr)

    return captured.out

def compare_output(monkeypatch, capsys, argv, output):
    assert run(monkeypatch, capsys, argv) == output


@pytest.mark.parametrize(
    "argv,output",
    [
        ([join("tests", "data", "hello_world.py")], ""),
        (
            [join("tests", "data", "custom_nonpep_assign.py")],
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style [default]\n"
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:2: E225 missing whitespace around operator [default]\n",
        ),
        (
            [join("tests", "data", "custom_flake8_pylint.py")],
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
    ],
)
def test_single_file_stdout(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)

@pytest.mark.parametrize(
    "argv",
    [
        [join("tests", "data", "hello_world.py")],
        [join("tests", "data", "custom_nonpep_assign.py")],
    ],
)
def test_single_file_stdout_json(monkeypatch, capsys, argv):
    out = run(monkeypatch, capsys, argv + ["--json"])
    result = json.loads(out)
    assert result

@pytest.mark.parametrize(
    "argv,output",
    [
        (
            [
                join("tests", "data", "hello_world.py"),
                join("tests", "data", "custom_nonpep_assign.py"),
            ],
            "****************** custom_nonpep_assign.py\n"
            f'{join("tests", "data", "custom_nonpep_assign.py")}:1:0: C0103 Constant name "a" doesn\'t conform to UPPER_CASE naming style [default]\n'
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:2: E225 missing whitespace around operator [default]\n",
        ),
        (
            [
                join("tests", "data", "custom_nonpep_assign.py"),
                join("tests", "data", "custom_flake8_pylint.py"),
            ],
            "****************** custom_nonpep_assign.py\n"
            f'{join("tests", "data", "custom_nonpep_assign.py")}:1:0: C0103 Constant name "a" doesn\'t conform to UPPER_CASE naming style [default]\n'
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:2: E225 missing whitespace around operator [default]\n"
            "****************** custom_flake8_pylint.py\n"
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            [
                join("tests", "data", "custom_flake8_pylint.py"),
                join("tests", "data", "custom_nonpep_assign.py"),
            ],
            "****************** custom_flake8_pylint.py\n"
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n"
            "****************** custom_nonpep_assign.py\n"
            f'{join("tests", "data", "custom_nonpep_assign.py")}:1:0: C0103 Constant name "a" doesn\'t conform to UPPER_CASE naming style [default]\n'
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:2: E225 missing whitespace around operator [default]\n",
        ),
    ],
)
def test_multiple_files_stdout(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize(
    "argv,output",
    [
        (
            [
                join("tests", "data", "custom_nonpep_assign.py"),
                join("tests", "data", "custom_flake8_pylint_config.py"),
            ],
            "****************** custom_nonpep_assign.py\n"
            f'{join("tests", "data", "custom_nonpep_assign.py")}:1:0: C0103 Constant name "a" doesn\'t conform to UPPER_CASE naming style [default]\n'
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:2: E225 missing whitespace around operator [default]\n"
            "****************** custom_flake8_pylint_config.py\n"
            f"{join('tests', 'data', 'custom_flake8_pylint_config.py')}:1:0: C0114 Missing module docstring [in-file]\n"
            f'{join("tests", "data", "custom_flake8_pylint_config.py")}:3:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint_config.py')}:4:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            [
                join("tests", "data", "custom_flake8_pylint_config.py"),
                join("tests", "data", "custom_nonpep_assign.py"),
            ],
            "****************** custom_flake8_pylint_config.py\n"
            f"{join('tests', 'data', 'custom_flake8_pylint_config.py')}:1:0: C0114 Missing module docstring [in-file]\n"
            f'{join("tests", "data", "custom_flake8_pylint_config.py")}:3:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint_config.py')}:4:5: F841 local variable 'a' is assigned to but never used [default]\n"
            "****************** custom_nonpep_assign.py\n"
            f'{join("tests", "data", "custom_nonpep_assign.py")}:1:0: C0103 Constant name "a" doesn\'t conform to UPPER_CASE naming style [default]\n'
            f"{join('tests', 'data', 'custom_nonpep_assign.py')}:1:2: E225 missing whitespace around operator [default]\n",
        ),
    ],
)
def test_different_configs(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize(
    "argv,output",
    [
        (
            f"{join('tests', 'data', 'custom_flake8_pylint.py')} -o pylint=--enable=missing-module-docstring".split(),
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:1:0: C0114 Missing module docstring [cmd]\n"
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            f"{join('tests', 'data', 'custom_flake8_pylint.py')} -o=pylint=--enable=missing-module-docstring".split(),
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:1:0: C0114 Missing module docstring [cmd]\n"
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            f"{join('tests', 'data', 'custom_flake8_pylint.py')} --option pylint=--enable=missing-module-docstring".split(),
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:1:0: C0114 Missing module docstring [cmd]\n"
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            f"{join('tests', 'data', 'custom_flake8_pylint.py')} --option=pylint=--enable=missing-module-docstring".split(),
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:1:0: C0114 Missing module docstring [cmd]\n"
            f'{join("tests", "data", "custom_flake8_pylint.py")}:1:0: C0104 Disallowed name "foo" [default]\n'
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            f"{join('tests', 'data', 'custom_flake8_pylint.py')} "
            "-o pylint=--enable=missing-module-docstring "
            "-o pylint=--disable=C0104".split(),
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:1:0: C0114 Missing module docstring [cmd]\n"
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
        (
            [
                join("tests", "data", "custom_flake8_pylint.py"),
                "-o",
                "pylint=--enable=missing-module-docstring pylint=--disable=C0104",
            ],
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:1:0: C0114 Missing module docstring [cmd]\n"
            f"{join('tests', 'data', 'custom_flake8_pylint.py')}:2:5: F841 local variable 'a' is assigned to but never used [default]\n",
        ),
    ],
)
def test_passing_cmd_args(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize("argv,output", [([join("tests", "data", "custom_swap.py")], "")])
def test_ib111_week(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)



from edulint.options import Option, T, UnionT, ImmutableT
from typing import Optional, Generic
from dataclasses import dataclass


@dataclass(frozen=True)
class Arg(Generic[T]):
    option: Option
    val: T


UnprocessedArg = Arg[Optional[str]]
ProcessedArg = Arg[UnionT]
ImmutableArg = Arg[ImmutableT]



from typing import Tuple, List, Union

from astroid import nodes  # type: ignore

from edulint.linting.analyses.antiunify import (
    antiunify,
    cprint,  # noqa: F401
    get_sub_variant,
    contains_avar,
)
from edulint.linting.analyses.variable_modification import VarEventType
from edulint.linting.analyses.data_dependency import vars_in, is_changed_between
from edulint.linting.analyses.cfg.utils import get_cfg_loc
from edulint.linting.checkers.utils import (
    is_parents_elif,
    get_lines_between,
    get_statements_count,
    get_token_count,
    has_else_block,
    is_negation,
    are_identical,
    new_node,
)
from edulint.linting.checkers.duplication.utils import (
    Fixed,
    length_mismatch,
    type_mismatch,
    called_aunify_var,
    assignment_to_aunify_var,
    is_duplication_candidate,
    saves_enough_tokens,
    get_loop_repetitions,
    to_node,
    to_parent,
    get_common_parent,
)

### constructive helper functions


def extract_from_elif(
    node: nodes.If, result: List[nodes.If] = None
) -> Tuple[bool, List[List[nodes.NodeNG]]]:
    """
    returns True iff elifs end with else
    """

    def count_nested_ifs(ns: List[nodes.NodeNG]) -> int:
        if len(ns) != 1 or not isinstance(ns[0], nodes.If):
            return 0

        if_ = ns[0]
        return max(count_nested_ifs(if_.body), count_nested_ifs(if_.orelse)) + 1

    result = [node] if result is None else result
    if has_else_block(node):
        return True, result

    current = node
    nested_count = count_nested_ifs(node.body)
    while current.has_elif_block():
        elif_ = current.orelse[0]
        result.append(elif_)
        current = elif_

    if nested_count == 0 or nested_count >= len(result):
        return has_else_block(current), result
    return True, result[:-nested_count]


def get_bodies(ifs: List[nodes.If]) -> List[List[nodes.NodeNG]]:
    result = []
    for i, if_ in enumerate(ifs):
        result.append(if_.body)
        if i == len(ifs) - 1:
            result.append(if_.orelse)
    return result


def extract_from_siblings(node: nodes.If, seq_ifs: List[nodes.NodeNG]) -> None:
    sibling = node.next_sibling()
    while sibling is not None and isinstance(sibling, nodes.If):
        new: List[nodes.NodeNG] = []
        if not extract_from_elif(sibling, new):
            return
        seq_ifs.append(sibling)
        seq_ifs.extend(new)
        sibling = sibling.next_sibling()


def create_ifs(tests: List[nodes.NodeNG]) -> Tuple[nodes.If, List[nodes.NodeNG]]:
    root = new_node(nodes.If)
    if_ = root
    if_bodies = []
    for i, test in enumerate(tests):
        if_.test = test
        if_bodies.append(if_.body)
        if i != len(tests) - 1:
            elif_ = new_node(nodes.If)
            if_.orelse = [elif_]
            # elif_.parent = if_
            if_ = elif_
        else:
            if_bodies.append(if_.orelse)
    return root, if_bodies


### testing helper functions


def is_one_of_parents_ifs(node: nodes.If) -> bool:
    parent = node.parent
    if not isinstance(parent, nodes.If):
        return False

    while isinstance(parent.parent, nodes.If):
        parent = parent.parent

    _ends_with_else, ifs = extract_from_elif(parent)
    if_bodies = get_bodies(ifs)

    return all(any(isinstance(n, nodes.If) for n in body) for body in if_bodies)


def contains_other_duplication(core, avars) -> bool:
    parent = get_common_parent(avars)
    if parent is None:
        body = get_sub_variant(core, 0)
    elif isinstance(parent, (nodes.List, nodes.Tuple, nodes.Dict, nodes.Set)):
        body = list(get_sub_variant(parent, 0).get_children())
    else:
        return False

    if len(body) < 3:
        return False

    for end, to_aunify in get_loop_repetitions(body):
        if not is_duplication_candidate(to_aunify):
            continue

        result = antiunify(
            to_aunify,
            stop_on=lambda avars: length_mismatch(avars) or type_mismatch(avars),
        )
        if result is not None:
            return True

    return False


def test_variables_change(tests, core, avars):
    vars = vars_in(tests, {VarEventType.READ})
    first_loc = tests[0].cfg_loc
    avars_locs = [get_cfg_loc(to_parent(avar)).node.sub_locs for avar in avars]
    return any(
        is_changed_between(var, first_loc, avar_locs)
        for var in vars.keys()
        for avar_locs in avars_locs
    )


def check_enabled(message_ids: Union[str, List[str]]):
    if isinstance(message_ids, str):
        message_ids = [message_ids]

    def middle(func):
        def inner(checker, *args, **kwargs):
            if not any(checker.linter.is_message_enabled(mi) for mi in message_ids):
                return None
            result = func(*args, **kwargs)
            if result is None:
                return result

            if len(message_ids) == 1:
                return Fixed(message_ids[0], *result)
            symbol, *result = result
            return Fixed(symbol, *result)

        inner.__name__ = func.__name__
        return inner

    return middle


### identical code before/after branch


def identical_before_after_branch(checker, ends_with_else: bool, ifs: List[nodes.If]) -> bool:

    def get_stmts_difference(branches: List[nodes.NodeNG], forward: bool) -> int:
        reference = branches[0]
        compare = branches[1:]
        for i in range(min(map(len, branches))):
            for branch in compare:
                index = i if forward else -i - 1
                if reference[index].as_string() != branch[index].as_string():
                    return i
        return i + 1

    def add_message(
        branches: List[nodes.NodeNG],
        stmts_difference: int,
        defect_node: nodes.NodeNG,
        forward: bool = True,
    ) -> None:
        reference = branches[0]
        first = reference[0 if forward else -stmts_difference]
        last = reference[stmts_difference - 1 if forward else -1]
        lines_difference = get_lines_between(first, last, including_last=True)

        checker.add_message(
            "identical-if-branches-part",
            node=defect_node,
            args=(lines_difference, "before" if forward else "after"),
        )

    if not ends_with_else:
        return False

    branches = get_bodies(ifs)

    any_message = False
    same_prefix_len = get_stmts_difference(branches, forward=True)
    if same_prefix_len >= 1:
        if all(same_prefix_len == len(b) for b in branches):
            checker.add_message("identical-if-branches", node=ifs[0])
            return True

        add_message(branches, same_prefix_len, ifs[0], forward=True)
        if any(same_prefix_len == len(b) for b in branches):
            return True

        any_message = True

    same_suffix_len = get_stmts_difference(branches, forward=False)
    if same_suffix_len >= 1:
        # allow wip early returns
        if same_suffix_len == 1 and isinstance(branches[0][-1], nodes.Return):
            return any_message
        defect_node = branches[0][-1].parent

        add_message(branches, same_suffix_len, defect_node, forward=False)
        any_message = True
    return any_message


### identical sequential ifs


def identical_seq_ifs(checker, ends_with_else: bool, ifs: List[nodes.If]) -> bool:

    def same_ifs_count(seq_ifs: List[nodes.NodeNG], start: int) -> int:
        reference = seq_ifs[start].body
        for i in range(start + 1, len(seq_ifs)):
            # do not suggest join of elif and sibling
            if seq_ifs[start].parent not in seq_ifs[i].node_ancestors():
                return i - start

            compared = seq_ifs[i].body
            if len(reference) != len(compared):
                return i - start
            for j in range(len(reference)):
                if reference[j].as_string() != compared[j].as_string():
                    return i - start
        return len(seq_ifs) - start

    if ends_with_else:
        return False

    prev_sibling = ifs[0].previous_sibling()
    if isinstance(prev_sibling, nodes.If) and not extract_from_elif(prev_sibling)[0]:
        return False

    extract_from_siblings(ifs[0], ifs)

    if len(ifs) == 1:
        return False

    i = 0
    last = None
    while i < len(ifs) - 1:
        count = same_ifs_count(ifs, i)
        if count > 1:
            first = ifs[i]
            assert isinstance(ifs[i + count - 1], nodes.If)
            last = ifs[i + count - 1].body[-1]

            checker.add_message(
                "identical-seq-ifs",
                line=first.fromlineno,
                col_offset=first.col_offset,
                end_lineno=last.tolineno,
                end_col_offset=last.end_col_offset,
                args=(count,),
            )
        i += count

    return last is not None


### if to restructured


def restructure_twisted_ifs(tests, inner_if: nodes.If, avars):
    if len(tests) > 1 or len(inner_if.orelse) == 0:
        return None

    outer_test = tests[0]
    inner_test = inner_if.test

    pos_pos = get_sub_variant(inner_if.body, 0)
    pos_neg = get_sub_variant(inner_if.orelse, 0)
    neg_pos = get_sub_variant(inner_if.body, 1)
    neg_neg = get_sub_variant(inner_if.orelse, 1)

    # positive and negative branches are twisted
    if not (
        are_identical(pos_pos, neg_neg)
        and are_identical(pos_neg, neg_pos)
        and not contains_avar(inner_test, avars)
    ) and not (
        # branches are correctly structured, but the inner condition is negated
        are_identical(pos_pos, neg_pos)
        and are_identical(pos_neg, neg_neg)
        and is_negation(
            get_sub_variant(inner_test, 0), get_sub_variant(inner_test, 1), negated_rt=False
        )
    ):
        return None

    pp_test = nodes.BoolOp(op="and")
    pp_test.values = [outer_test, inner_test]

    neg_outer_test = nodes.UnaryOp(op="not")
    neg_outer_test.operand = outer_test
    neg_inner_test = nodes.UnaryOp(op="not")
    neg_inner_test.operand = inner_test

    nn_test = nodes.BoolOp(op="and")
    nn_test.values = [neg_outer_test, neg_inner_test]

    test = nodes.BoolOp(op="or")
    test.values = [pp_test, nn_test]

    if_ = nodes.If()
    if_.test = test
    if_.body = inner_if.sub_locs[0].node.body
    if_.orelse = inner_if.sub_locs[0].node.orelse

    return if_


@check_enabled("similar-if-to-untwisted")
def get_fixed_by_restructuring_twisted(tests, core, avars):
    if len(tests) > 1 or len(core) != 1 or not isinstance(core[0], nodes.If):
        return None

    inner_if = core[0]
    if len(inner_if.orelse) == 0:
        return None

    outer_test = tests[0]
    inner_test = inner_if.test

    pos_pos = get_sub_variant(inner_if.body, 0)
    pos_neg = get_sub_variant(inner_if.orelse, 0)
    neg_pos = get_sub_variant(inner_if.body, 1)
    neg_neg = get_sub_variant(inner_if.orelse, 1)

    # positive and negative branches are twisted
    if not (
        are_identical(pos_pos, neg_neg)
        and are_identical(pos_neg, neg_pos)
        and not contains_avar(inner_test, avars)
    ) and not (
        # branches are correctly structured, but the inner condition is negated
        are_identical(pos_pos, neg_pos)
        and are_identical(pos_neg, neg_neg)
        and is_negation(
            get_sub_variant(inner_test, 0), get_sub_variant(inner_test, 1), negated_rt=False
        )
    ):
        return None

    inner_test_variant = get_sub_variant(inner_test, 0)
    pp_test = new_node(nodes.BoolOp, op="and", values=[outer_test, inner_test_variant])

    neg_outer_test = new_node(nodes.UnaryOp, op="not", operand=outer_test)
    neg_inner_test = new_node(nodes.UnaryOp, op="not", operand=inner_test_variant)

    nn_test = new_node(nodes.BoolOp, op="and", values=[neg_outer_test, neg_inner_test])

    test = new_node(nodes.BoolOp, op="or", values=[pp_test, nn_test])

    if_ = new_node(
        nodes.If,
        test=test,
        body=inner_if.sub_locs[0].node.body,
        orelse=inner_if.sub_locs[0].node.orelse,
    )

    return (
        get_token_count(if_),
        get_statements_count(if_, include_defs=False, include_name_main=False),
        (test.as_string(),),
    )


### if to use


@check_enabled("similar-if-to-use")
def get_fixed_by_if_to_use(tests, core, avars):
    if len(tests) > 1 or len(avars) > 1:
        return None

    avar = avars[0]
    if (
        not isinstance(avar.parent, nodes.Const)
        or not isinstance(avar.subs[0], bool)
        or not isinstance(avar.subs[1], bool)
    ):
        return None

    # do not repeat simplifiable if
    if isinstance(avar.parent.parent, (nodes.Return, nodes.Assign)):
        return None

    return (
        get_token_count(core) - 1 + get_token_count(tests[0]),
        get_statements_count(core, include_defs=False, include_name_main=False),
        (),
    )


### if to ternary

COMPLEX_EXPRESSION_TYPES = (nodes.BinOp,)
# SIMPLE_EXPRESSION_TYPES = (nodes.AugAssign, nodes.Call, nodes.BoolOp, nodes.Compare)


def is_part_of_complex_expression(avars) -> bool:
    for avar in avars:
        has_parent_call = False
        parent = to_parent(avar)
        while parent is not None:
            if isinstance(parent, nodes.Call):
                if has_parent_call or len(parent.args) > 1:
                    return True
                has_parent_call = True

            # if not isinstance(parent, SIMPLE_EXPRESSION_TYPES):
            if isinstance(parent, COMPLEX_EXPRESSION_TYPES):
                return True

            if hasattr(parent, "cfg_loc"):
                break

            parent = parent.parent

    return False


@check_enabled("similar-if-to-expr")
def get_fixed_by_ternary(tests, core, avars):
    # the condition would get too complicated
    if len(tests) > 1 and any(isinstance(test, nodes.BoolOp) for test in tests):
        return None
    # too much place for error
    if len(avars) > 1 and not all(isinstance(avar.parent, nodes.Const) for avar in avars):
        return None
    # do not make complicated expressions even more complicated
    if is_part_of_complex_expression(avars):
        return None

    # generate exprs
    exprs = []
    for avar in avars:
        assert len(avar.subs) == len(tests) + 1
        expr = to_node(avar.subs[-1], avar)
        for test, avar_val in reversed(list(zip(tests, avar.subs))):
            new = new_node(nodes.IfExp, test=test, body=to_node(avar_val, avar), orelse=expr)
            expr = new

        exprs.append(expr)

    return (
        get_token_count(core) - len(avars) + get_token_count(exprs),  # subtract aunify vars
        get_statements_count(core, include_defs=False, include_name_main=True),
        (exprs[0].as_string(),),
    )


### if into block

HEADER_ATTRIBUTES = {
    nodes.For: ["target", "iter"],
    nodes.While: ["test"],
    nodes.If: ["test"],
    nodes.ExceptHandler: ["name", "type"],
    nodes.Try: [],
    nodes.With: ["items"],
}

BODY_ATTRIBUTES = {
    nodes.For: ["body", "orelse"],
    nodes.While: ["body", "orelse"],
    nodes.If: ["body", "orelse"],
    nodes.ExceptHandler: ["body"],
    nodes.Try: ["body", "handlers", "orelse", "finalbody"],
    nodes.With: ["body"],
}


def if_can_be_moved(core, avars):
    if type(core) not in HEADER_ATTRIBUTES.keys():
        return False

    # avar in header
    if any(contains_avar(getattr(core, attr), avars) for attr in HEADER_ATTRIBUTES[type(core)]):
        return False

    return (
        sum(contains_avar(getattr(core, attr), avars) for attr in BODY_ATTRIBUTES[type(core)]) == 1
    )


def get_fixed_by_moving_if_rec(tests, core, avars):
    if isinstance(core, list):
        if len(core) == 0:
            return []

        avar_indices = []
        for i, stmt in enumerate(core):
            if contains_avar(stmt, avars):
                avar_indices.append(i)

        assert len(avar_indices) > 0
        min_ = avar_indices[0]
        max_ = avar_indices[-1]

        if min_ == max_ and if_can_be_moved(core[min_], avars):
            root, lines = get_fixed_by_moving_if_rec(tests, core[min_], avars)
        else:
            new_body = core[min_ : max_ + 1]
            root, if_bodies = create_ifs(tests)
            for if_body in if_bodies:
                if_body.extend(new_body)
            lines = [
                (new_body[0].sub_locs[i].node.fromlineno, new_body[-1].sub_locs[i].node.tolineno)
                for i in range(len(new_body[0].sub_locs))
            ]
            lines = [f"{start}" if start == end else f"{start}-{end}" for start, end in lines]

        return core[:min_] + [root] + core[max_ + 1 :], lines

    assert contains_avar(core, avars) and if_can_be_moved(core, avars)
    new_core = new_node(type(core))

    for attr in HEADER_ATTRIBUTES[type(core)]:
        setattr(new_core, attr, getattr(core, attr))

    lines = []
    for attr in BODY_ATTRIBUTES[type(core)]:
        attrval = getattr(core, attr)
        if contains_avar(attrval, avars):
            new_body, sublines = get_fixed_by_moving_if_rec(tests, getattr(core, attr), avars)
            lines.extend(sublines)
        else:
            new_body = attrval
        setattr(new_core, attr, new_body)

    return new_core, lines


@check_enabled("similar-if-into-block")
def get_fixed_by_moving_if(tests, core, avars):
    # too restrictive -- the change may be before the avar but after the place
    # where the if would be inserted
    if (not isinstance(core, list) and not if_can_be_moved(core, avars)) or (
        isinstance(core, list) and not if_can_be_moved(core[0], avars)
    ):
        return None

    fixed, lines = get_fixed_by_moving_if_rec(tests, core, avars)
    return (
        get_token_count(fixed)
        + sum(
            get_token_count(v) if isinstance(v, nodes.NodeNG) else 0
            for avar in avars
            for v in avar.subs
        ),
        get_statements_count(fixed, include_defs=True, include_name_main=False),
        (", ".join(lines),),
    )


### if to variables


@check_enabled("similar-if-to-extracted")
def get_fixed_by_vars(tests, core, avars):
    root, if_bodies = create_ifs(tests)
    seen = {}
    for avar in avars:
        var_vals = tuple(avar.subs)
        varname = seen.get(var_vals, avar.name)

        if varname != avar.name:
            continue
        seen[var_vals] = avar.name

        for val, body in zip(var_vals, if_bodies):
            assign = new_node(nodes.Assign)
            assign.targets = [new_node(nodes.AssignName, name=avar.name)]
            assign.value = to_node(val, avar)
            body.append(assign)

    return (
        get_token_count(root) + get_token_count(core),
        get_statements_count(root, include_defs=False, include_name_main=True)
        + get_statements_count(core, include_defs=False, include_name_main=True),
        (),
    )


### control functions


def identical_blocks_in_if(checker, ends_with_else: bool, ifs: List[nodes.If]) -> bool:
    if identical_before_after_branch(checker, ends_with_else, ifs):
        return True

    return identical_seq_ifs(checker, ends_with_else, ifs)


def similar_blocks_in_if(checker, ends_with_else: bool, ifs: List[nodes.If]) -> bool:
    if not ends_with_else:
        return False

    # do not break up consistent ifs
    if is_one_of_parents_ifs(ifs[0]):
        return False

    if_bodies = get_bodies(ifs)
    assert len(if_bodies) >= 2
    result = antiunify(
        if_bodies,
        stop_on=lambda avars: length_mismatch(avars) or type_mismatch(avars),
        stop_on_after_renamed_identical=lambda avars: assignment_to_aunify_var(avars),
    )
    if result is None:
        return False
    core, avars = result

    if contains_other_duplication(core, avars):
        return False

    if len(avars) == 0:
        checker.add_message("identical-if-branches", node=ifs[0])
        return True

    tokens_before = get_token_count(ifs[0])
    stmts_before = get_statements_count(ifs[0], include_defs=False, include_name_main=True)

    tests = [if_.test for if_ in ifs]
    called_avar = called_aunify_var(avars)
    tvs_change = test_variables_change(tests, core, avars)

    for fix_function in (
        get_fixed_by_restructuring_twisted if not tvs_change else None,
        get_fixed_by_if_to_use if not tvs_change else None,
        get_fixed_by_moving_if if not tvs_change else None,
        get_fixed_by_ternary if not called_avar and not tvs_change else None,
        get_fixed_by_vars if not called_avar else None,
    ):
        if fix_function is None:
            continue

        suggestion = fix_function(checker, tests, core, avars)
        if suggestion is None:
            continue

        if not saves_enough_tokens(tokens_before, stmts_before, suggestion):
            continue

        message_id, _tokens, _statements, message_args = suggestion
        checker.add_message(message_id, node=ifs[0], args=message_args)
        return True

    return False


def duplicate_in_if(checker, node: nodes.If) -> Tuple[bool, bool]:
    if is_parents_elif(node):
        return False, False

    ends_with_else, ifs = extract_from_elif(node)
    if identical_blocks_in_if(checker, ends_with_else, ifs):
        return True, True

    return similar_blocks_in_if(checker, ends_with_else, ifs), False



import re

from astroid import nodes

from edulint.linting.analyses.cfg.utils import get_cfg_loc
from edulint.linting.analyses.variable_scope import ScopeListener
from edulint.linting.checkers.utils import is_builtin
from edulint.linting.analyses.var_events import (
    VarEventType,
    VarName,
    Variable,
    VarEvent,
    strip_to_name,
)


class VarModificationAnalysis:
    @staticmethod
    def collect(node: nodes.NodeNG):
        VarEventListener().visit(node)


class VarEventListener(ScopeListener[None]):
    NON_PURE_METHODS = re.compile(
        r"append|clear|extend|insert|pop|remove|reverse|sort|add|.*update|write"
    )

    def __init__(self):
        super().__init__()

    @staticmethod
    def may_be_unscoped(name: str, name_node: nodes.NodeNG, loc_node: nodes.NodeNG):
        if name.startswith("__"):
            return True
        if is_builtin(name_node, name):
            return True
        if (
            isinstance(name_node, nodes.Attribute)
            and isinstance(name_node.parent, nodes.Call)
            and name_node.parent.func == name_node
        ):
            return True

        for parent in name_node.node_ancestors():
            if parent == loc_node.parent:
                break
            if isinstance(parent, nodes.Call):  # and is_builtin(parent.func, name):
                return True

        return False

    def _add_var_event(self, name: VarName, name_node: nodes.NodeNG, action: VarEventType):
        loc = get_cfg_loc(name_node)
        scope = self._get_var_scope(name)
        # assert scope is not None or VarEventListener.may_be_unscoped(
        #     name, name_node, loc.node
        # ), f"but {name_node.as_string()}, {loc.node.as_string()}"
        if scope is not None:
            var = Variable(name, scope)
            loc.var_events.add(var, VarEvent(var, name_node, action))

    # @override
    def _init_var_in_scope(self, name: VarName, name_node: nodes.NodeNG, offset: int = 0) -> None:
        was_defined = name in self.stack[-1 + offset]
        super()._init_var_in_scope(name, name_node, offset)
        self._add_var_event(
            name, name_node, VarEventType.ASSIGN if not was_defined else VarEventType.REASSIGN
        )

    # @override
    def _del_var_from_scope(self, name: VarName, node: nodes.NodeNG):
        self._add_var_event(name, node, VarEventType.DELETE)
        super()._del_var_from_scope(name, node)

    # @override
    def _visit_assigned_to(self, node: nodes.NodeNG) -> None:
        super()._visit_assigned_to(node)

        stripped = strip_to_name(node)
        if stripped is None:
            return

        if isinstance(node, (nodes.AssignAttr, nodes.Subscript)):
            self._add_var_event(stripped.name, stripped, VarEventType.MODIFY)

    # @override
    def visit_augassign(self, node: nodes.AugAssign) -> None:
        stripped = strip_to_name(node.target)
        if stripped is not None:
            self._add_var_event(stripped.name, node, VarEventType.READ)

        super().visit_augassign(node)

    def visit_name(self, node: nodes.Name) -> None:
        self._add_var_event(node.name, node, VarEventType.READ)

    def visit_attribute(self, node: nodes.Attribute) -> None:
        stripped = strip_to_name(node)
        if (
            isinstance(node.parent, nodes.Call)
            and VarEventListener.NON_PURE_METHODS.match(node.attrname)
            and stripped is not None
        ):
            self._add_var_event(stripped.name, stripped, VarEventType.MODIFY)

        return self.visit_many(node.get_children())

    def visit_assignattr(self, node: nodes.AssignAttr) -> None:
        stripped = strip_to_name(node)
        if stripped is not None:
            self._add_var_event(stripped.name, stripped, VarEventType.MODIFY)

        return self.visit_many(node.get_children())

    def visit_delattr(self, node: nodes.DelAttr) -> None:
        stripped = strip_to_name(node)
        if stripped is None:
            return
        self._add_var_event(stripped.name, stripped, VarEventType.MODIFY)



# Some modifications are based on good ideas from Thonny
# https://raw.githubusercontent.com/thonny/thonny/master/thonny/plugins/pylint/
# tho_xpln > msg_xpln


def make_explanation_more_friedly(explanation: str) -> str:
    replace_prefixes = {
        "Used when an ": "It looks like the ",
        "Used when a ": "It looks like the ",
        "Used when ": "It looks like ",
        "Emitted when an ": "It looks like the ",
        "Emitted when a ": "It looks like the ",
        "Emitted when ": "It looks like ",
    }

    for prefix, replacement in replace_prefixes.items():
        if explanation.startswith(prefix):
            explanation = replacement + explanation[len(prefix) :]
            break

    return explanation



from typing import List

from astroid import nodes  # type: ignore

from edulint.linting.analyses.antiunify import (
    antiunify,
    core_as_string,
    new_node,
    cprint,  # noqa: F401
)
from edulint.linting.analyses.data_dependency import vars_in
from edulint.linting.checkers.utils import (
    get_statements_count,
    get_token_count,
    get_range_params,
    get_const_value,
)
from edulint.linting.checkers.duplication.utils import (
    Fixed,
    length_mismatch,
    type_mismatch,
    called_aunify_var,
    assignment_to_aunify_var,
    saves_enough_tokens,
    to_node,
    to_start_lines,
)


class NoSubseqToLoop(Exception):
    pass


### constructive helper functions


def to_range_args(sequence):
    start = None
    step = None
    previous = None
    for s in sequence:
        if not isinstance(s, int):
            return None
        if start is None:
            start = s
        elif step is None:
            step = s - start
        else:
            assert previous is not None
            if s - previous != step:
                return None
        previous = s

    return (start, previous + (1 if step > 0 else -1), step)


def to_range_node(range_args):
    start, stop, step = range_args

    range_ = new_node(nodes.Call, func=new_node(nodes.Name, name="range"))
    start_node = new_node(nodes.Const, value=start)
    stop_node = new_node(nodes.Const, value=stop)
    step_node = new_node(nodes.Const, value=step)
    if step != 1:
        range_.args = [start_node, stop_node, step_node]
    elif start != 0:
        range_.args = [start_node, stop_node]
    else:
        range_.args = [stop_node]
    return range_


### nice sequence helpers


def partition_by_type(sequence):
    result = []
    current = []

    for n in sequence:
        if len(current) == 0 or isinstance(n, type(current[-1])):
            current.append(n)
        else:
            result.append(current)
            current = [n]

    if len(current) > 0:
        result.append(current)

    return result


def to_const_sequence(sequence):
    result = []
    for v in sequence:
        c = get_const_value(v)
        if c is None:
            return None
        result.append(c)
    return result


def from_chars(avar, sequence):
    if not isinstance(avar.parent, nodes.Const):
        return None

    if not all(isinstance(s, str) and len(s) == 1 for s in sequence):
        return None

    min_char = min(sequence)

    sequence = [ord(s) - ord(min_char) for s in sequence]

    if ord(min_char) == 0:
        return sequence, avar

    # chr(ord(min_char) + ID)
    ord_call = new_node(
        nodes.Call,
        func=new_node(nodes.Name, name="ord"),
        args=[new_node(nodes.Const, value=min_char)],
    )

    binop = new_node(nodes.BinOp, op="+", left=ord_call, right=avar)

    chr_call = new_node(nodes.Call, func=new_node(nodes.Name, name="chr"), args=[binop])

    return sequence, chr_call


def iter_use_from_partition(partition):
    types = [type(p[0]) for p in partition]
    # a type repeats
    if len(types) != len(set(types)):
        return None

    exclusive = set(types) & {nodes.Name, nodes.Subscript, nodes.Attribute}
    # different exclusive types or an exclusive type multiple times
    if len(exclusive) > 1:
        return None

    type_groups = {type(p[0]): p for p in partition}
    for t in (nodes.Const, nodes.Name, nodes.Subscript, nodes.Attribute, nodes.BinOp):
        type_groups[t] = type_groups.get(t, [])

    # multiple constants or no binops
    if len(type_groups[nodes.Const]) > 1 or len(type_groups[nodes.BinOp]) == 0:
        return None

    if len(exclusive) == 1:
        exclusive_type = next(iter(exclusive))
        # multiple values for an exclusive type
        if len(type_groups[exclusive_type]) > 1:
            return None
        exclusive_value = type_groups[exclusive_type][0]
    else:
        exclusive_type = None
        exclusive_value = None

    if not any(
        ts
        in (
            [nodes.Const, exclusive_type, nodes.BinOp],
            [exclusive_type, nodes.BinOp],
            [nodes.Const, nodes.BinOp],
        )
        for ts in (types, list(reversed(types)))
    ):
        return None

    binop_core, bionp_avars = antiunify(type_groups[nodes.BinOp])
    assert isinstance(binop_core, nodes.BinOp)
    # all same binops and binops differing in multiple places break niceness
    if len(bionp_avars) != 1 or called_aunify_var(bionp_avars):
        return None
    binop_avar = bionp_avars[0]

    if len(type_groups[nodes.Const]) == 1:
        const_value = type_groups[nodes.Const][0]
    else:
        const_value = None

    # no child is related to the shared value
    if (
        exclusive_value is not None
        and binop_core.right.as_string() != exclusive_value.as_string()
        and binop_core.left.as_string() != exclusive_value.as_string()
    ) or (
        exclusive_value is None
        and const_value is not None
        and core_as_string(binop_core.right) != const_value.as_string()
        and core_as_string(binop_core.left) != const_value.as_string()
    ):
        return None

    if const_value is not None:
        const_nums = [0]
    else:
        const_nums = []

    if exclusive_value is not None:
        if binop_core.op == "+" or (
            binop_core.op == "-" and exclusive_value.as_string() == binop_core.left.as_string()
        ):
            exclusive_nums = [0]
        elif binop_core.op == "*" or (
            binop_core.op in ("/", "//", "%")
            and exclusive_value.as_string() == binop_core.left.as_string()
        ):
            exclusive_nums = [1]
        else:
            exclusive_nums = []
    else:
        exclusive_nums = []

    if isinstance(binop_avar.subs[0], nodes.NodeNG):
        binop_result = iter_use_from_partition(partition_by_type(binop_avar.subs))
        if binop_result is None:
            return None
        binop_nums, sub_binop_use = binop_result
        if binop_core.left == binop_avar.parent:
            binop_core.left = sub_binop_use
        else:
            binop_core.right = sub_binop_use
    else:
        binop_nums = binop_avar.subs

    dct = {
        nodes.Const: const_nums,
        exclusive_type: exclusive_nums,
        nodes.BinOp: binop_nums,
    }

    return [n for t in types for n in dct[t]], binop_core


def to_iter_use(avar):
    sequence = list(avar.subs)
    use = avar

    const_sequence = to_const_sequence(sequence)
    if const_sequence is not None:
        sequence = const_sequence

    from_chars_result = from_chars(avar, sequence)
    if from_chars_result is not None:
        sequence, use = from_chars_result

    range_args = to_range_args(sequence)
    if range_args is not None:
        return range_args, use

    partition = partition_by_type(sequence)
    # single type present => use values directly, if different
    if len(partition) == 1:
        assert not any(isinstance(v, nodes.NodeNG) for v in sequence)
        if len(sequence) != len(set(sequence)):  # some value is repeated
            return None
        return sequence, use

    result = iter_use_from_partition(partition)
    if result is None:
        return None
    range_nums, use = result
    range_args = to_range_args(range_nums)
    if range_args is None:
        return None
    return range_args, use


def consolidate_ranges(ranges):
    if len(ranges) == 1:
        range_args, use = ranges[0]
        return [to_range_node(range_args)], [use]

    uses = []
    for (start, stop, step), use in ranges:
        if step != 1:
            use = new_node(nodes.BinOp, op="*", left=use, right=new_node(nodes.Const, value=step))
        if start != 0:
            use = new_node(nodes.BinOp, op="+", left=use, right=new_node(nodes.Const, value=start))
        uses.append(use)

    start, stop, step = 0, (stop - start - 1) // step + 1, 1
    return [to_range_node((start, stop, step))], uses


def get_nice_iters(avars, to_aunify):
    sequences = [avar.subs for avar in avars]
    if len(sequences) == 0:
        range_node = new_node(
            nodes.Call,
            func=new_node(nodes.Name, name="range"),
            args=[new_node(nodes.Const, value=len(to_aunify))],
        )
        return [range_node], {}

    iter_uses = []
    for avar in avars:
        result = to_iter_use(avar)
        if result is None:
            return None
        iter, use = result
        iter_uses.append((iter, use))

    ranges = [(iter, use) for iter, use in iter_uses if isinstance(iter, tuple)]

    # disallow mixing ranges with collections
    # TODO maybe too strict?
    if len(ranges) != 0 and len(ranges) != len(iter_uses):
        return None

    if len(ranges) == 0:
        str_iters = {
            tuple(to_node(n, avars[i]).as_string() for n in iter)
            for i, (iter, _use) in enumerate(iter_uses)
        }
        if len(str_iters) != 1:
            return None
        some_iter, use = iter_uses[0]
        if isinstance(use, nodes.Call) and use.func.name == "chr":
            start = use.args[0].left.args[0].value
            some_iter = [chr(ord(start) + v) for v in some_iter]

        collection = new_node(nodes.List, elts=[to_node(n, avars[0]) for n in some_iter])
        return [collection], [use for _iter, use in iter_uses]

    if len({r[0] for r in ranges}) > 2:
        raise NoSubseqToLoop

    return consolidate_ranges(ranges)


### fixers


def get_iter(iters):
    return iters[0]


def get_target(_avars, _iters):
    return new_node(nodes.AssignName, name="i")


def get_fixed_by_merging_with_parent_loop(to_aunify, core, avars):
    parent = to_aunify[0][0].parent
    if len(avars) > 0 or not isinstance(parent, nodes.For):
        return None

    first = to_aunify[0][0]
    last = to_aunify[-1][-1]

    if first != parent.body[0] or last != parent.body[-1]:
        return None

    range_params = get_range_params(parent.iter)
    if range_params is None:
        return None

    start, stop, step = range_params
    if get_const_value(start) != 0 or get_const_value(step) != 1:
        return None

    used_vars = vars_in([n for ns in to_aunify for n in ns])
    target = parent.target
    # TODO can be weakened -- use div to get i's original value
    if (
        not isinstance(target, nodes.AssignName)
        or (target.name, parent.scope()) in used_vars.keys()
    ):
        return None

    const_stop = get_const_value(stop)
    new_iter = (
        f"{const_stop * len(to_aunify)}"
        if isinstance(stop, nodes.Const)
        else f"{stop.as_string()} * {len(to_aunify)}"
    )
    return Fixed(
        "similar-block-to-loop-merge",
        get_token_count(core),
        get_statements_count(core, include_defs=False, include_name_main=True),
        (to_start_lines(to_aunify), new_iter),
    )


def get_fixed_by_loop(to_aunify, core, avars):
    result = get_nice_iters(avars, to_aunify)
    if result is None:
        return None
    iters, uses = result
    assert len(iters) == 1

    iterable = get_iter(iters)
    for_ = new_node(nodes.For, iter=iterable, target=get_target(avars, iters), body=core)

    symbol = (
        "similar-block-to-loop-range"
        if isinstance(iterable, nodes.Call)
        else "similar-block-to-loop-collection"
    )

    return Fixed(
        symbol,
        get_token_count(for_),
        get_statements_count(for_, include_defs=False, include_name_main=True),
        (
            len(to_aunify),
            get_statements_count(to_aunify[0], include_defs=False, include_name_main=True),
            iterable.as_string(),
        ),
    )


### control function


def similar_to_loop(self, to_aunify: List[List[nodes.NodeNG]]) -> bool:
    if max(len(to_aunify), len(to_aunify[0])) <= 2:  # TODO parametrize?
        return False

    result = antiunify(
        to_aunify,
        stop_on=lambda avars: length_mismatch(avars)
        or type_mismatch(
            avars,
            allowed_mismatches=[
                {nodes.Const, nodes.BinOp, t}
                for t in (nodes.Name, nodes.Subscript, nodes.Attribute)
            ],
        )
        or called_aunify_var(avars),
        stop_on_after_renamed_identical=lambda avars: assignment_to_aunify_var(avars),
    )
    if result is None:
        return False
    core, avars = result

    tokens_before = sum(get_token_count(node) for node in to_aunify)
    stmts_before = sum(
        get_statements_count(node, include_defs=False, include_name_main=True) for node in to_aunify
    )

    fixed_by_merge = get_fixed_by_merging_with_parent_loop(to_aunify, core, avars)
    if fixed_by_merge is not None:
        fixed = fixed_by_merge
    else:
        try:
            fixed = get_fixed_by_loop(to_aunify, core, avars)
        except NoSubseqToLoop:
            return True

        if fixed is None:
            return False

    if not saves_enough_tokens(tokens_before, stmts_before, fixed):
        return False

    message_id, _tokens, _statements, message_args = fixed

    first = to_aunify[0][0]
    last = to_aunify[-1][-1]
    self.add_message(
        message_id,
        line=first.fromlineno,
        col_offset=first.col_offset,
        end_lineno=last.tolineno,
        end_col_offset=last.col_offset,
        args=message_args,
    )
    return True



import os

import pytest

from edulint.options import Option
from edulint.config.arg import Arg
from edulint.config.language_translations import Translation, parse_lang_file
from test_utils import lazy_problem, apply_and_lint, get_tests_path

def get_lang_translations_path(filename):
    return get_tests_path(os.path.join("language_translations", filename))

@pytest.mark.parametrize("filename,expected_output", [
    ("R6609-pylint-translation-by-symbol.toml", {"use-augmented-assign": Translation("Use '{} {}= {}'", extracts={})}),
])
def test_parse_lang_file(filename: str, expected_output):
    assert parse_lang_file(get_lang_translations_path(filename)) == expected_output

@pytest.mark.parametrize("checked_filename,translations_filename,expected_output", [
    ("z202817-zkouska.py", "R6609-pylint-translation-by-symbol.toml", [
        lazy_problem().set_line(10).set_text("Use 'num //= 10'"),
        lazy_problem().set_line(172)
        .set_text("do not compare types, for exact checks use `is` / `is not`, for instance checks use `isinstance()`"),
        lazy_problem().set_line(196).set_text("Unnecessary pass statement"),
    ]),
    ("z202817-zkouska.py", "R6609-pylint-translation-by-code.toml", [
        lazy_problem().set_line(10).set_text("Use 'num //= 10'"),
        lazy_problem().set_line(172)
        .set_text("do not compare types, for exact checks use `is` / `is not`, for instance checks use `isinstance()`"),
        lazy_problem().set_line(196).set_text("Unnecessary pass statement"),
    ]),
    ("z202817-zkouska.py", "E721-flake8-translation-by-code.toml", [
        lazy_problem().set_line(10).set_text("Use augmented assignment: 'num //= 10'"),
        lazy_problem().set_line(172).set_text("Use isinstance"),
        lazy_problem().set_line(196).set_text("Unnecessary pass statement"),
    ]),
    ("064fe05979.py", "E712-flake8-translation-tweaked.toml", [
        lazy_problem().set_line(41).set_text("Instead of comparison with False use 'if not cond'"),
    ]),
    ("064fe05979.py", "E712-flake8-translation-swapped.toml", [
        lazy_problem().set_line(41).set_text("Use 'if not cond' instead of comparison with False"),
    ]),
    ("064fe05979.py", "E712-flake8-translation-only-second.toml", [
        lazy_problem().set_line(41).set_text("Use 'if not cond'"),
    ]),
    ("z202817-zkouska.py", "many.toml", [
        lazy_problem().set_line(10).set_text("Use 'num //= 10'"),
        lazy_problem().set_line(172).set_text("Use isinstance"),
        lazy_problem().set_line(196).set_text("You shall not pass"),
    ]),
])
def test_translations(checked_filename, translations_filename, expected_output):
    apply_and_lint(
        get_tests_path(checked_filename),
        [
            Arg(Option.PYLINT, "--enable=R6609,W0107"),
            Arg(Option.FLAKE8, "--extend-select=E721,E712"),
            Arg(Option.LANGUAGE_FILE, get_lang_translations_path(translations_filename))
        ],
        expected_output
    )

@pytest.mark.parametrize("checked_filename,translations_filename,expected_output", [
    ("z202817-zkouska.py", "config-file-direct-translations-many.toml", [
        lazy_problem().set_line(10).set_text("Use 'num //= 10'"),
        lazy_problem().set_line(172).set_text("Use isinstance"),
        lazy_problem().set_line(196).set_text("You shall not pass"),
    ]),
    ("z202817-zkouska.py", "config-file-referenced-translations-many.toml", [
        lazy_problem().set_line(10).set_text("Use 'num //= 10'"),
        lazy_problem().set_line(172).set_text("Use isinstance"),
        lazy_problem().set_line(196).set_text("You shall not pass"),
    ]),
])
def test_translations_in_config_file(checked_filename, translations_filename, expected_output):
    apply_and_lint(
        get_tests_path(checked_filename),
        [Arg(Option.CONFIG_FILE, get_lang_translations_path(translations_filename))],
        expected_output
    )



from typing import Any, Optional, Union, Dict
from loguru import logger

from edulint.options import Option


def print_invalid_type_message(option: Union[Option, str], val: Any) -> None:
    logger.warning(
        "invalid value type {type} of value {val} for option {option}",
        type=type(val),
        val=val,
        option=option.to_name() if isinstance(option, Option) else option,
    )


def config_file_val_to_str(option: Union[Option, str], val: Any) -> Optional[str]:
    if isinstance(val, int):
        return str(val)
    if isinstance(val, str):
        return val
    if isinstance(val, list):
        return ",".join(val)
    if isinstance(val, tuple):
        key, value = val
        value = config_file_val_to_str(option, value)
        return f"--{key}={value}" if value is not None else None
    print_invalid_type_message(option, val)
    return None


def add_enabled(enabler_name: str, enablers: Dict[str, str], option: Option, val: str):
    def extract(linter_option: str):
        if val.startswith(linter_option):
            for symbol in val[len(linter_option) :].split(","):
                enablers[symbol] = enabler_name
        return {}

    if option == Option.PYLINT:
        extract("--enable=")
    elif option == Option.FLAKE8:
        extract("--select=")
        extract("--extend-select=")
    if option == Option.IGNORE_INFILE_CONFIG_FOR:
        enablers["EDL001"] = enabler_name



"""
Provides a function to generate and display the control flow graph of a given module.
"""

# adapted from https://github.com/pyta-uoft/pyta/blob/4c858623549e24a49fea7aef9c8ec7c20c836bd6/python_ta/cfg/cfg_generator.py

import importlib.util
import os.path
import sys
from typing import Any, Dict, Optional, Set

import graphviz
from astroid import nodes
from astroid.builder import AstroidBuilder

from edulint.linting.analyses.cfg.graph import CFGBlock, ControlFlowGraph
from edulint.linting.analyses.cfg.visitor import CFGVisitor

GRAPH_OPTIONS = {
    "format": "dot",
    "node_attr": {"shape": "box", "fontname": "Courier New"},
}
SUBGRAPH_OPTIONS = {"fontname": "Courier New"}


def generate_cfg(
    mod: str = "", auto_open: bool = False, visitor_options: Optional[Dict[str, Any]] = None
) -> None:
    """Generate a control flow graph for the given module.

    Supported Options:
      - "separate-condition-blocks": bool
            This option specifies whether the test condition of an if statement gets merged with any
            preceding statements or placed in a new block. By default, it will merge them.
      - "functions": list[str]
            This option specifies whether to restrict the creation of cfgs to just top-level
            function definitions or methods provided in this list. By default, it will create the
            cfg for the entire file.

    Args:
        mod (str): The path to the module. `mod` can either be the path of a file (must have `.py`
            extension) or have no argument (generates a CFG for the Python file from which this
            function is called).
        auto_open (bool): Automatically open the graph in your browser.
        visitor_options (dict): An options dict to configure how the cfgs are generated.
    """
    _generate(mod=mod, auto_open=auto_open, visitor_options=visitor_options)


def _generate(
    mod: str = "", auto_open: bool = False, visitor_options: Optional[Dict[str, Any]] = None
) -> None:
    """Generate a control flow graph for the given module.

    `mod` can either be:
      - the path of a file (must have `.py` extension).
      - no argument -- generate a CFG for the Python file from which this function is called.
    """
    # Generate a control flow graph for the given file
    abs_path = _get_valid_file_path(mod)
    # Print an error message if the file is not valid and early return
    if abs_path is None:  # _get_valid_file_path returns None in case of invalid file
        return

    file_name = os.path.splitext(os.path.basename(abs_path))[0]
    module = AstroidBuilder().file_build(abs_path)
    visitor = CFGVisitor(options=visitor_options)
    module.accept(visitor)

    _display(visitor.cfgs, file_name, auto_open=auto_open)


def _get_valid_file_path(mod: str = "") -> Optional[str]:
    """Return the valid absolute path of `mod`, a path to the target file."""
    # Allow call to check with empty args
    if mod == "":
        m = sys.modules["__main__"]
        spec = importlib.util.spec_from_file_location(m.__name__, m.__file__)
        mod = spec.origin
    # Enforce the API to only except `mod` type as str
    elif not isinstance(mod, str):
        print(
            "No CFG generated. Input to check, `{}`, has invalid type, must be a string.".format(
                mod
            )
        )
        return

    # At this point, `mod` is of type str
    if not os.path.isfile(mod):
        # `mod` is not a file so print an error message
        print("Could not find the file called, `{}`\n".format(mod))
        return

    # `mod` may be a relative path to a valid file so return its absolute path
    return os.path.abspath(mod)


def _display(
    cfgs: Dict[nodes.NodeNG, ControlFlowGraph], filename: str, auto_open: bool = False
) -> None:
    graph = graphviz.Digraph(name=filename, **GRAPH_OPTIONS)
    for node, cfg in cfgs.items():
        if isinstance(node, nodes.Module):
            subgraph_label = "__main__"
        elif isinstance(node, nodes.FunctionDef):
            scope_parent = node.scope().parent
            subgraph_label = node.name
            # Update the label to the qualified name if it is a method
            if isinstance(scope_parent, nodes.ClassDef):
                subgraph_label = scope_parent.name + "." + subgraph_label
        else:
            continue
        with graph.subgraph(name=f"cluster_{cfg.cfg_id}") as c:
            visited = set()
            _visit(cfg.start, c, visited, cfg.end)
            for block in cfg.unreachable_blocks:
                _visit(block, c, visited, cfg.end)
            c.attr(label=subgraph_label, **SUBGRAPH_OPTIONS)

    graph.render(filename, view=auto_open)


def display_from_block(block: CFGBlock, filename: str, auto_open: bool = False) -> None:
    graph = graphviz.Digraph(name=filename, **GRAPH_OPTIONS)

    with graph.subgraph(name="cluster") as c:
        visited = set()
        _visit(block, c, visited, None)
        c.attr(**SUBGRAPH_OPTIONS)

    graph.render(filename, view=auto_open, cleanup=True)


def _visit(block: CFGBlock, graph: graphviz.Digraph, visited: Set[int], end: CFGBlock) -> None:
    """
    Visit a CFGBlock and add it to the control flow graph.
    """
    node_id = f"{graph.name}_{block.id}"
    if node_id in visited:
        return

    label = "\n".join([loc.node.as_string() for loc in block.locs]) + "\n"
    # Need to escape backslashes explicitly.
    label = label.replace("\\", "\\\\")
    # \l is used for left alignment.
    label = label.replace("\n", "\\l")

    fill_color = "grey93" if not block.reachable else "white"
    # Change the fill colour if block is the end of the cfg
    fill_color = "black" if block == end else fill_color

    graph.node(node_id, label=label, fillcolor=fill_color, style="filled")
    visited.add(node_id)

    for edge in block.successors:
        if edge.label is not None:
            graph.edge(node_id, f"{graph.name}_{edge.target.id}", str(edge.label))
        else:
            graph.edge(node_id, f"{graph.name}_{edge.target.id}")
        _visit(edge.target, graph, visited, end)



from typing import Dict, List, TypeVar, Union, Optional, Iterator

from astroid import nodes

from edulint.linting.analyses.var_events import VarName, ScopeNode
from edulint.linting.checkers.utils import BaseVisitor, is_builtin


T = TypeVar("T")


def in_new_scope(func):
    def inner(self, *args, **kwargs):
        self.stack.append({})
        result = func(self, *args, **kwargs)
        self.stack.pop()
        return result

    return inner


class UnknowableLocalsException(Exception):
    pass


class ScopeListener(BaseVisitor[T]):
    def __init__(self):
        self.stack: List[Dict[VarName, ScopeNode]] = []
        super().__init__()

    def _get_var_scope_index(self, name: VarName) -> Optional[int]:
        for i, scope in enumerate(reversed(self.stack), 1):
            if name in scope:
                return -i
        return None

    def _get_var_scope(self, name: VarName) -> Optional[Dict[str, nodes.NodeNG]]:
        scope_index = self._get_var_scope_index(name)
        if scope_index is None:
            return None
        return self.stack[scope_index][name]

    def _init_var_in_scope(self, name: VarName, name_node: nodes.NodeNG, offset: int = 0) -> None:
        if name not in self.stack[-1 + offset]:
            self.stack[-1 + offset][name] = name_node.scope()

    def _del_var_from_scope(self, name: VarName, _node: nodes.NodeNG):
        scope_index = self._get_var_scope_index(name)
        if scope_index is None:
            return

        self.stack[scope_index].pop(name)

    @staticmethod
    def statements_before_definitions(node: Union[nodes.Module, nodes.FunctionDef]):
        sooner = []
        later = []
        for child in node.body:
            if not isinstance(child, (nodes.FunctionDef, nodes.ClassDef)):
                sooner.append(child)
            else:
                later.append(child)
        return sooner + later

    @in_new_scope
    def visit_module(self, node: nodes.Module) -> T:
        return self.visit_many(ScopeListener.statements_before_definitions(node))

    @in_new_scope
    def visit_functiondef(self, node: nodes.FunctionDef) -> T:
        self._init_var_in_scope(node.name, node, offset=-1)

        for arg in node.args.arguments:
            self._init_var_in_scope(arg.name, arg)

        for arg in node.args.kwonlyargs:
            self._init_var_in_scope(arg.name, arg)

        if node.args.vararg is not None:
            self._init_var_in_scope(node.args.vararg, node.args)

        return self.visit_many(ScopeListener.statements_before_definitions(node))

    @in_new_scope
    def visit_classdef(self, node: nodes.ClassDef) -> T:
        self._init_var_in_scope(node.name, node, offset=-1)

        return self.visit_many(node.get_children())

    @in_new_scope
    def visit_lambda(self, node: nodes.Lambda) -> T:
        for arg in node.args.arguments:
            self._init_var_in_scope(arg.name, arg)

        return self.visit(node.body)

    def _visit_generator(
        self, node: Union[nodes.GeneratorExp, nodes.ListComp, nodes.DictComp, nodes.SetComp]
    ):
        self.visit_many(node.generators)
        if isinstance(node, nodes.DictComp):
            self.visit(node.key)
            return self.visit(node.value)
        return self.visit(node.elt)

    @in_new_scope
    def visit_generatorexp(self, node: nodes.GeneratorExp) -> T:
        return self._visit_generator(node)

    @in_new_scope
    def visit_listcomp(self, node: nodes.ListComp) -> T:
        return self._visit_generator(node)

    @in_new_scope
    def visit_dictcomp(self, node: nodes.DictComp) -> T:
        return self._visit_generator(node)

    @in_new_scope
    def visit_setcomp(self, node: nodes.SetComp) -> T:
        return self._visit_generator(node)

    def visit_global(self, node: nodes.Global) -> T:
        for name in node.names:
            self.stack[-1][name] = node.root().scope()

        return self.default

    def visit_nonlocal(self, node: nodes.Nonlocal) -> T:
        for name in node.names:
            self.stack[-1][name] = node.scope().parent.scope()

        return self.default

    def _visit_assigned_to(self, node: nodes.NodeNG) -> None:
        if isinstance(node, nodes.AssignName):
            self._init_var_in_scope(node.name, node)
        return self.visit_many(node.get_children())

    def _names_from_tuple(self, targets: List[nodes.NodeNG]) -> Iterator[nodes.NodeNG]:
        for target in targets:
            if not isinstance(target, nodes.Tuple):
                yield target
            else:
                yield from self._names_from_tuple(target.elts)

    def visit_assignname(self, node: nodes.AssignName) -> T:
        self._visit_assigned_to(node)
        return self.default

    def visit_assign(self, node: nodes.Assign) -> T:
        self.visit(node.value)

        for target in self._names_from_tuple(node.targets):
            self._visit_assigned_to(target)

        return self.default

    def visit_augassign(self, node: nodes.AugAssign) -> T:
        self.visit(node.value)
        self._visit_assigned_to(node.target)
        return self.default

    def visit_annassign(self, node: nodes.AnnAssign) -> T:
        if node.value is not None:
            self.visit(node.value)
        self._visit_assigned_to(node.target)
        return self.default

    def visit_delname(self, node: nodes.DelName) -> T:
        self._del_var_from_scope(node.name, node)
        return self.default

    def visit_import(self, node: nodes.Import) -> T:
        for module, alias in node.names:
            name = module if alias is None else alias
            self._init_var_in_scope(name, node)
        return self.default

    def visit_importfrom(self, node: nodes.ImportFrom) -> T:
        return self.visit_import(node)

    def visit_call(self, node: nodes.Call) -> T:
        if node.func.as_string() in ("locals", "globals") and is_builtin(node.func):
            raise UnknowableLocalsException(
                f"call to {node.func.as_string()} on line {node.fromlineno} in {node.root().file}; "
                " unable to determine variable scopes. This may disable some detectors."
            )
        return self.visit_many(node.get_children())

    def visit_for(self, node: nodes.For) -> T:
        return self.visit_many(node.get_children())



from typing import List, Set, Dict
import re
from loguru import logger
import os

from edulint.linters import Linter
from edulint.options import Option
from edulint.linting.problem import Problem


CONFIG_PATTERNS = {
    Linter.FLAKE8: [
        re.compile(r"(.*)#\s*(noqa)", re.IGNORECASE),
        re.compile(r"(.*)#\s*(flake8: *noqa)", re.IGNORECASE),
    ],
    Linter.PYLINT: [
        re.compile(r"(.*)#\s*(pylint:\s*disable)", re.IGNORECASE),
        re.compile(r"(.*)#\s*(pylint:\s*enable)", re.IGNORECASE),
    ],
    Linter.EDULINT: [],  # handled in config
}


def to_file_paths(files_or_dirs, prefix=""):
    for path in files_or_dirs:
        path = os.path.join(prefix, path)
        if not os.path.isdir(path):
            if path.endswith(".py"):
                yield path
        else:
            yield from to_file_paths(os.listdir(path), path)


def report_infile_config(
    files_or_dirs: List[str], ignore_infile: Set[str], enablers: Dict[str, str]
) -> List[Problem]:
    if "all" in ignore_infile:
        ignore_infile = (ignore_infile - {"all"}) | {linter.to_name() for linter in Linter}

    patterns = []
    for ignore_infile_linter in ignore_infile:
        linter = Linter.safe_from_name(ignore_infile_linter)
        if linter is None:
            logger.warning(
                "invalid value '{val}' for option {option}",
                val=ignore_infile_linter,
                option=Option.IGNORE_INFILE_CONFIG_FOR.to_name(),
            )
            continue
        patterns.extend(CONFIG_PATTERNS[linter])

    results = []
    ib111_re = re.compile(r".*from\s+ib111\s+import", re.IGNORECASE)
    for file_path in to_file_paths(files_or_dirs):
        with open(file_path, encoding="utf-8") as f:
            for i, line in enumerate(f, 1):
                for pattern in patterns:
                    match = pattern.match(line)
                    if match and ("noqa" not in match.group(2).lower() or not ib111_re.match(line)):
                        results.append(
                            Problem(
                                source=Linter.EDULINT,
                                enabled_by=enablers.get("EDL001"),
                                path=file_path,
                                line=i,
                                column=len(match.group(1)),
                                code="EDL001",
                                text=f"Forbidden magic comment '{match.group(2)}'",
                                end_line=i,
                                end_column=len(line),
                            )
                        )
    return results



import edulint
from edulint.options import Option
from edulint.option_parses import OptionParse, get_option_parses
from edulint.config.config import get_config_many, get_cmd_args, ImmutableConfig
from edulint.config.language_translations import LangTranslations
from edulint.linting.problem import Problem
from edulint.linting.linting import lint_many, sort, EduLintLinterFailedException
from edulint.versions.version_checker import PackageInfoManager
from edulint.explanations import update_explanations, get_explanations
from typing import List, Dict, Tuple, Any
import argparse
import os
import sys
import json
from threading import Thread
from loguru import logger


def setup_logger() -> None:
    logger.remove()
    logger.add(
        sys.stderr,
        level="WARNING",
        format="<level>{name}</level>: {message}",
        diagnose=False,
        backtrace=False,
        catch=False,
    )


def format_options_help(option_parses: Dict[Option, OptionParse]) -> str:
    def extract_line(words: List[str], max_len: int, start_i: int) -> Tuple[str, int]:
        result = [words[start_i]]
        current_len = len(words[start_i])
        for i in range(start_i + 1, len(words)):
            next_len = current_len + 1 + len(words[i])
            if next_len > max_len:
                return " ".join(result), i
            result.append(words[i])
            current_len += 1 + len(words[i])
        return " ".join(result), len(words)

    result = []
    for op in option_parses.values():
        words = (op.help_).split(" ")
        i = 0
        while i < len(words):
            if i == 0:
                start = op.option.to_name() + "  "
            else:
                start = " " * 4
            line, i = extract_line(words, 80 - 25 - len(start), i)
            result.append(start + line)
    return "\n".join(result) + "\n"


def setup_argparse(option_parses: Dict[Option, OptionParse]) -> argparse.Namespace:
    main_parser = argparse.ArgumentParser(prog="edulint")

    shared_options_parser = argparse.ArgumentParser(add_help=False)
    shared_options_parser.add_argument(
        "--json", action="store_true", help="should output problems in json format"
    )
    shared_options_parser.add_argument(
        "--disable-version-check",
        action="store_true",
        default=False,
        help="EduLint checks for a newer version at most once per hour. If newer version is available in pip, "
        "it will print a message to stderr. Specifying this flag disables the check completely.",
    )
    shared_options_parser.add_argument(
        "--disable-explanations-update",
        action="store_true",
        default=False,
        help="EduLint periodically updates explanations. Specifying this flag disables the updates.",
    )

    subparsers = main_parser.add_subparsers(dest="command")
    subparsers.required = True

    check_parser = subparsers.add_parser(
        "check",
        description="Lints provided code.",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[shared_options_parser],
    )
    check_parser.add_argument(
        "-o",
        "--option",
        metavar="OPTION",
        dest="options",
        default=[],
        action="append",
        help=format_options_help(option_parses),
    )
    check_parser.add_argument(
        "files_or_dirs",
        metavar="FILE-OR-DIRECTORY",
        nargs="+",
        help="the file(s) or directory(ies) to lint",
    )

    explain_parser = subparsers.add_parser(
        "explain", description="Explains check by message ID", parents=[shared_options_parser]
    )
    explain_parser.add_argument(
        "message_ids",
        metavar="MESSAGE-ID",
        nargs="+",
        help="message id (e.g., E0001); use 'all' as a message id to get all explanations",
    )

    _version_parser = subparsers.add_parser(
        "version", description="Shows installed EduLint version", parents=[shared_options_parser]
    )

    return main_parser.parse_args()


def to_json(
    configs: List[Tuple[List[str], ImmutableConfig, LangTranslations]], problems: List[Problem]
) -> str:
    def config_to_json(obj: Any) -> str:
        if isinstance(obj, ImmutableConfig):
            return {arg.option.to_name(): arg.val for arg in obj.config}
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable")

    config_json = json.dumps(
        [config for _files, config, _translation in configs], default=config_to_json
    )
    problems_json = Problem.schema().dumps(problems, indent=2, many=True)
    return f'{{"configs": {config_json}, "problems": {problems_json}}}'


def check_for_updates(is_check_disabled: bool = False):
    if is_check_disabled:
        return

    Thread(target=_update_check).start()


def _update_check():
    python_executable = sys.executable or (
        "python" if os.name == "nt" else "python3"
    )  # nt is Windows
    version_ttl = 600
    if PackageInfoManager.is_update_waiting("edulint", ttl=version_ttl):
        try:
            logger.warning(
                f"Update for EduLint ({PackageInfoManager.get_latest_version('edulint', version_ttl)}) is available. "
                f"You can upgrade using `{python_executable} -m pip install --upgrade --user edulint`"
            )
        except ValueError as e:
            if not (len(e.args) > 0 and e.args[0] == "I/O operation on closed file."):
                raise e


def check_code(args, option_parses):
    cmd_args = get_cmd_args(args)

    file_configs = get_config_many(args.files_or_dirs, cmd_args, option_parses=option_parses)

    try:
        results = lint_many(file_configs)
    except (TimeoutError, json.decoder.JSONDecodeError, EduLintLinterFailedException):
        return 2

    sorted_results = sort(args.files_or_dirs, results)
    checks_single_file = len(args.files_or_dirs) == 1 and not os.path.isdir(args.files_or_dirs[0])

    if args.json:
        print(to_json(file_configs, sorted_results))
    else:
        prev_problem = None
        for problem in sorted_results:
            if not checks_single_file and (
                prev_problem is None or prev_problem.path != problem.path
            ):
                print(f"****************** {os.path.basename(problem.path)}")
                prev_problem = problem

            print(problem)

    return 0 if len(sorted_results) == 0 else 1


def explain_messages(args):
    explanations = get_explanations()
    if any(id_.lower() == "all" for id_ in args.message_ids):
        message_ids = explanations.keys()
    else:
        message_ids = args.message_ids

    mid_expls = {}
    for mid in message_ids:
        expl = explanations.get(mid)
        if expl is None:
            logger.warning(f"Message {mid} does not have an explanation")
        elif args.json:
            mid_expls[mid] = expl
        else:
            if len(args.message_ids) > 1:
                if mid != args.message_ids[0]:
                    print("\n")
                print(f"## {mid}")
            if expl.get("why", "").strip():
                print("### Why is it a problem?")
                print(expl["why"].strip())
                print()
            if expl.get("examples", "").strip():
                print("### How to solve it?")
                print(expl["examples"].strip())

    if args.json:
        data = {"explanations": mid_expls}
        print(json.dumps(data, indent=2))

    return 0


@logger.catch
def main() -> int:
    setup_logger()
    option_parses = get_option_parses()
    args = setup_argparse(option_parses)

    check_for_updates(args.disable_version_check)
    update_explanations(
        args.disable_explanations_update
    )  # get_explanations also can trigger update, but we're not calling it anywhere else.

    if args.command == "check":
        return check_code(args, option_parses)
    if args.command == "explain":
        return explain_messages(args)
    if args.command == "version":
        print(f"edulint version {edulint.__version__}")
        return 0
    assert False, "unreachable, but " + args.command



from dataclasses import dataclass
from typing import Dict, List, Any, Optional, Callable
from enum import Enum, auto
from edulint.options import Option, UnionT, T, DEFAULT_CONFIG
from edulint.linters import Linter


class TakesVal(Enum):
    YES = auto()
    NO = auto()
    OPTIONAL = auto()


class MultivaluedEnum(Enum):
    def __new__(cls, *args: Any, **kwds: Any) -> Any:
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj


class Type(MultivaluedEnum):
    @staticmethod
    def _to_bool_val(val: Optional[str]) -> bool:
        return val is None or val in ("true", "on")

    @staticmethod
    def _to_list_val(val: Optional[str]) -> List[str]:
        return [val] if val is not None else []

    @staticmethod
    def _from_comma_separated_to_list_val(val: Optional[str]) -> List[str]:
        return val.split(",") if val else []

    @staticmethod
    def _to_str_val(val: Optional[str]) -> Optional[str]:
        return val

    @staticmethod
    def _to_int_val(val: Optional[str]) -> Optional[int]:
        return int(val) if val is not None and val.isdecimal() else None

    def __init__(self, _: Enum, convert: Callable[[Optional[str]], UnionT]):
        self.convert: Callable[[Optional[str]], UnionT] = convert.__func__  # type: ignore

    def __call__(self, arg: Optional[str]) -> UnionT:
        return self.convert(arg)

    BOOL = (auto(), _to_bool_val)
    LIST = (auto(), _to_list_val)
    COMMA_SEPARATED_LIST = (auto, _from_comma_separated_to_list_val)
    STR = (auto(), _to_str_val)
    INT = (auto(), _to_int_val)


class Combine(MultivaluedEnum):
    @staticmethod
    def _keep_right_combine(_lt: T, rt: T) -> T:
        return rt

    @staticmethod
    def _append_combine(lt: List[T], rt: T) -> List[T]:
        return lt + [rt]

    @staticmethod
    def _extend_combine(lt: List[T], rt: List[T]) -> List[T]:
        return lt + rt

    def __init__(self, _: Enum, combine: Callable[[UnionT, UnionT], UnionT]):
        self.combine: Callable[[UnionT, UnionT], UnionT] = combine.__func__  # type: ignore

    def __call__(self, lt: UnionT, rt: UnionT) -> UnionT:
        return self.combine(lt, rt)

    REPLACE = (auto(), _keep_right_combine)
    EXTEND = (auto(), _extend_combine)


@dataclass
class OptionParse:
    option: Option
    help_: str
    takes_val: TakesVal
    default: UnionT
    convert: Type
    combine: Combine


OPTIONS: List[OptionParse] = [
    OptionParse(
        Option.CONFIG_FILE,
        "config file to use for the linting (packaged name, local path or remote)",
        TakesVal.YES,
        DEFAULT_CONFIG,
        Type.STR,
        Combine.REPLACE,
    ),
    OptionParse(
        Option.PYLINT,
        "arguments to be passed to pylint (formatted to be passed through command line)",
        TakesVal.YES,
        [],
        Type.LIST,
        Combine.EXTEND,
    ),
    OptionParse(
        Option.FLAKE8,
        "arguments to be passed to flake8 (formatted to be passed through command line)",
        TakesVal.YES,
        [],
        Type.LIST,
        Combine.EXTEND,
    ),
    OptionParse(
        Option.ALLOWED_ONECHAR_NAMES,
        "only listed characters are allowed to be variable names of length one",
        TakesVal.YES,
        None,
        Type.STR,
        Combine.REPLACE,
    ),
    OptionParse(
        Option.NO_FLAKE8, "turn off flake8", TakesVal.NO, False, Type.BOOL, Combine.REPLACE
    ),
    OptionParse(
        Option.IGNORE_INFILE_CONFIG_FOR,
        "warns about infile supressions (like # noqa) for given linters, "
        f"valid values are {', '.join(linter.to_name() for linter in Linter)} and all, "
        f"using values {Linter.EDULINT} and all is recommended only from the command line, "
        f"if values {Linter.EDULINT} or all are set in-file or in the config file specified "
        "in-file, all other in-file configuration (possibly including the config from the "
        "config file) is ignored",
        TakesVal.YES,
        [],
        Type.COMMA_SEPARATED_LIST,
        Combine.REPLACE,
    ),
    OptionParse(
        Option.EXPORT_GROUPS,
        "sets which groups are to be advertised",
        TakesVal.YES,
        [],
        Type.COMMA_SEPARATED_LIST,
        Combine.REPLACE,
    ),
    OptionParse(
        Option.SET_GROUPS,
        "sets which groups are to be applied",
        TakesVal.YES,
        [],
        Type.COMMA_SEPARATED_LIST,
        Combine.REPLACE,
    ),
    OptionParse(
        Option.DISALLOWED_BUILTIN_NAMES,
        "listed built-in names are forbidden (if none are specified, all builtin names are disallowed)",
        TakesVal.YES,
        [],
        Type.COMMA_SEPARATED_LIST,
        Combine.REPLACE,
    ),
    OptionParse(
        Option.LANGUAGE_FILE,
        "file with translations for linter messages",
        TakesVal.YES,
        None,
        Type.STR,
        Combine.REPLACE,
    ),
]

OPTION_SETS_LABEL = "translations"
LANG_TRANSLATIONS_LABEL = "language-translations"
LANG_TRANSLATED_EXTRACTS_LABEL = "translated_extracts"
DEFAULT_ENABLER_LABEL = "default-enabler-name"


def get_option_parses() -> Dict[Option, OptionParse]:
    return {parse.option: parse for parse in OPTIONS}



from typing import Dict
import os
from threading import Thread
from pathlib import Path
from typing import Optional, Dict
import dbm
import time
import random

import tomli
import requests
from platformdirs import PlatformDirs
from loguru import logger


SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))
EXPLANATIONS_PIP_DISTRIBUTED_FILEPATH = os.path.join(SCRIPT_PATH, "explanations.toml")

EDULINT_LOCAL_DATA_FOLDER = PlatformDirs(appname="edulint").user_data_dir
EXPLANATIONS_ONLINE_DISTRIBUTED_FILEPATH = os.path.join(EDULINT_LOCAL_DATA_FOLDER, "explanations_online.toml")
EXPLANATIONS_DBM = os.path.join(EDULINT_LOCAL_DATA_FOLDER, "explanations.dbm")
GITHUB_URL = "https://raw.githubusercontent.com/GiraffeReversed/edulint/main/edulint/explanations.toml"

try:
    Path(EDULINT_LOCAL_DATA_FOLDER).mkdir(parents=True, exist_ok=True)
except Exception as e:
    logger.warning("Failed to create local directory for edulints internal files. {e}")

# --- Read explanations
def get_explanations(disable_explanations_update: bool = False) -> Dict[str, Dict[str, str]]:
    update_explanations(disable_explanations_update)  # This is async. Its results won't be ready during the lifetime of this function.

    updated_explanations = _load_updated_explanations()
    if updated_explanations:
        return updated_explanations

    with open(EXPLANATIONS_PIP_DISTRIBUTED_FILEPATH, "rb") as f:
        parsed_toml = tomli.load(f)
    return parsed_toml


def _load_updated_explanations() -> Optional[Dict[str, str]]:
    path_for_online_file = Path(EXPLANATIONS_ONLINE_DISTRIBUTED_FILEPATH)
    if not path_for_online_file.exists():
        return
    try:
        with open(EXPLANATIONS_PIP_DISTRIBUTED_FILEPATH, "rb") as f:
            return tomli.load(f)
    except:
        logger.warning("Updated explanations seem to be corrupted. Falling back to those distributed with the pip package.")


# --- Update explanation
def current_timestamp() -> int:
    return int(time.time())


def update_explanations(disable_explanations_update: bool = False):
    if disable_explanations_update:
        return
    Thread(target=_thread_update_explanations).start()


def _thread_update_explanations(ttl: int = 600): 
    try:
        if not Path(EXPLANATIONS_DBM).exists():
            time.sleep(random.randint(1,30)/1000)  # Small random delay to lower the chance for file creation race condition 
        with dbm.open(EXPLANATIONS_DBM, 'c') as db:
            if current_timestamp() < int(db.get('last_update', b'0')) + ttl:
                return
            db['last_update'] = str(current_timestamp())  # We're counting any update attempt, not just the succesfull ones
        logger.debug(f"Succesfully wrote current time to {EXPLANATIONS_DBM}.")

        resp = requests.get(GITHUB_URL, timeout=3)
        if resp.status_code != 200:
            return

        with open(EXPLANATIONS_ONLINE_DISTRIBUTED_FILEPATH, "w", encoding="utf8") as f:
            f.write(resp.text)
    except Exception as e:
        logger.debug(f"Update of explanations failed with {e}. This can ocassionaly hapen for some processes due to race condition.")

# todo: add test to load explanantions (toml parser is sensitive to small mistakes)

if __name__ == "__main__":
    get_explanations()



# combined from
# https://github.com/pyta-uoft/pyta/blob/683505e2a910c2a252739094593406e4e0f29a85/python_ta/checkers/invalid_for_target_checker.py
# https://github.com/pyta-uoft/pyta/blob/683505e2a910c2a252739094593406e4e0f29a85/python_ta/checkers/one_iteration_checker.py
# https://github.com/pyta-uoft/pyta/blob/683505e2a910c2a252739094593406e4e0f29a85/python_ta/checkers/shadowing_in_comprehension_checker.py
# https://github.com/pyta-uoft/pyta/blob/683505e2a910c2a252739094593406e4e0f29a85/python_ta/checkers/top_level_code_checker.py


from typing import TYPE_CHECKING, Union, List

import re

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.base import UpperCaseStyle
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter

from edulint.linting.checkers.utils import is_main_block


class InvalidForTargetChecker(BaseChecker):
    # name is the same as file name but without _checker part
    name = "invalid_for_target"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9984": (
            'For loop or comprehension variable "%s" should not be a part of a larger object.',
            "invalid-for-target",
            "Used when you have an index variable in a for loop or comprehension"
            "that is in subscript or object attribute form",
        )
    }
    # this is important so that your checker is executed before others
    priority = -1

    INVALID_TARGETS = (nodes.Subscript, nodes.AssignAttr)

    @only_required_for_messages("invalid-for-target")
    def visit_for(self, node: nodes.For) -> None:
        invalid_for_targets = node.target.nodes_of_class(self.INVALID_TARGETS)
        for target in invalid_for_targets:
            self.add_message("invalid-for-target", node=target, args=target.as_string())

    @only_required_for_messages("invalid-for-target")
    def visit_comprehension(self, node: nodes.Comprehension) -> None:
        invalid_for_targets = node.target.nodes_of_class(self.INVALID_TARGETS)
        for target in invalid_for_targets:
            self.add_message("invalid-for-target", node=target, args=target.as_string())


class OneIterationChecker(BaseChecker):
    # name is the same as file name but without _checker part
    name = "one_iteration"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9996": (
            "This loop will only ever run for one iteration",
            "one-iteration",
            "Reported when the loop body always ends the loop in its first iteration "
            '(e.g., by returning or using the "break" keyword).',
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    # pass in message symbol as a parameter of only_required_for_messages
    @only_required_for_messages("one-iteration")
    def visit_for(self, node: nodes.For) -> None:
        if self._check_one_iteration(node):
            self.add_message("one-iteration", node=node)

    @only_required_for_messages("one-iteration")
    def visit_while(self, node: nodes.While) -> None:
        if self._check_one_iteration(node):
            self.add_message("one-iteration", node=node)

    def _check_one_iteration(self, node: Union[nodes.For, nodes.While]) -> bool:
        """Return whether the given loop is guaranteed to stop after one iteration.
        More precisely, Returns False if there exists a direct predecessor
        block `p` to the start of the loop block `s` such that the
        first statement in `p` is a child node of <node> and that there exists a
        path from `s` to `p.
        Note: For `while` loops, 'start of the loop block' refers to the block with
        the test condition (or the first of the blocks that make up test condition).
        For `for` loops, it refers to the block with the assignment target.
        """
        start = node.target if isinstance(node, nodes.For) else node
        if not hasattr(start, "cfg_loc"):
            return False

        preds = start.cfg_loc.block.predecessors

        if preds == []:
            return False

        for pred in preds:
            stmt = pred.source.locs[0].node
            if node.parent_of(stmt) and pred.source.reachable:
                if isinstance(node, nodes.For) and stmt is node.iter:
                    continue
                return False
        return True


class ShadowingInComprehensionChecker(BaseChecker):
    name = "shadowing_in_comprehension"
    msgs = {
        "E9988": (
            "Comprehension variable '%s' shadows a variable in an outer scope",
            "shadowing-in-comprehension",
            "Used when there is shadowing inside a comprehension",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    @only_required_for_messages("shadowing-in-comprehension")
    def visit_comprehension(self, node: nodes.Comprehension) -> None:
        if isinstance(node.target, nodes.Tuple):
            for target in node.target.elts:
                if (
                    isinstance(target, nodes.Name)
                    and target.name in node.parent.frame().locals
                    and target.name != "_"
                ):
                    args = target.name
                    self.add_message("shadowing-in-comprehension", node=target, args=args)
        elif isinstance(node.target, nodes.AssignName):
            if node.target.name in node.parent.frame().locals and node.target.name != "_":
                args = node.target.name
                self.add_message("shadowing-in-comprehension", node=node.target, args=args)


class TopLevelCodeChecker(BaseChecker):
    name = "top_level_code"
    msgs = {
        "E9992": (
            "Forbidden top-level code found on line %s",
            "forbidden-top-level-code",
            "Used when you write top-level code that is not allowed. "
            "The allowed top-level code includes imports, definitions, and assignments.",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    @only_required_for_messages("forbidden-top-level-code")
    def visit_module(self, node: nodes.Module) -> None:
        for statement in node.body:
            if not (
                _is_import(statement)
                or _is_definition(statement)
                or _is_assignment(statement)
                or is_main_block(statement)
            ):
                self.add_message("forbidden-top-level-code", node=statement, args=statement.lineno)


# Helper functions
def _is_import(statement: nodes.NodeNG) -> bool:
    """
    Return whether or not <statement> is an Import or an ImportFrom.
    """
    return isinstance(statement, (nodes.Import, nodes.ImportFrom))


def _is_definition(statement: nodes.NodeNG) -> bool:
    """
    Return whether or not <statement> is a function definition or a class definition.
    """
    return isinstance(statement, (nodes.FunctionDef, nodes.ClassDef))


def _is_constant_assignment(statement: nodes.NodeNG) -> bool:
    """
    Return whether or not <statement> is a constant assignment.
    """
    if not isinstance(statement, nodes.Assign):
        return False

    names: List[str] = []
    for target in statement.targets:
        names.extend(node.name for node in target.nodes_of_class(nodes.AssignName, nodes.Name))

    return all(re.match(UpperCaseStyle.CONST_NAME_RGX, name) for name in names)


def _is_assignment(statement: nodes.NodeNG) -> bool:
    return isinstance(statement, nodes.Assign)


def register(linter: "PyLinter") -> None:
    linter.register_checker(InvalidForTargetChecker(linter))
    linter.register_checker(OneIterationChecker(linter))
    linter.register_checker(ShadowingInComprehensionChecker(linter))
    linter.register_checker(TopLevelCodeChecker(linter))



import pytest
from pathlib import Path
from edulint.linters import Linter
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from test_utils import lazy_problem, apply_and_lint, create_apply_and_lint, just_lint, get_tests_path
from typing import List


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        ("hello_world.py", [], []),
        (
            "z202817-zkouska.py",
            [],
            [
                lazy_problem().set_code("R6609").set_line(10),
                lazy_problem().set_code("R6303").set_line(42),
                lazy_problem().set_code("R6205").set_line(82),
                lazy_problem().set_code("R6307").set_line(171),
                lazy_problem().set_code("W0107").set_line(196),
            ],
        ),
        (
            "z202817-zkouska.py",
            [Arg(Option.PYLINT, "--enable=C0115")],
            [
                lazy_problem().set_code("R6609").set_line(10),
                lazy_problem().set_code("R6303").set_line(42),
                lazy_problem().set_code("R6205").set_line(82),
                lazy_problem().set_code("C0115").set_line(122),
                lazy_problem().set_code("C0115").set_line(128),
                lazy_problem().set_code("R6307").set_line(171),
                lazy_problem().set_code("W0107").set_line(196),
            ],
        ),
        (
            "z202817-zkouska.py",
            [Arg(Option.PYLINT, "--enable=C0115"), Arg(Option.PYLINT, "--disable=W0107,R6609")],
            [
                lazy_problem().set_code("R6303").set_line(42),
                lazy_problem().set_code("R6205").set_line(82),
                lazy_problem().set_code("C0115").set_line(122),
                lazy_problem().set_code("C0115").set_line(128),
                lazy_problem().set_code("R6307").set_line(171),
            ],
        ),
        (
            "z202817-zkouska.py",
            [Arg(Option.CONFIG_FILE, "empty"), Arg(Option.PYLINT, "--enable=R6609")],
            [lazy_problem().set_code("R6609").set_line(10)]
        ),
        (
            "002814-p1_trapezoid.py",
            [],
            [
                lazy_problem().set_code("F401").set_line(1),
                lazy_problem().set_code("F401").set_line(1),
                lazy_problem().set_code("E501").set_line(1),
                lazy_problem().set_code("W293").set_line(19),
                lazy_problem().set_code("E303").set_line(22),
            ],
        ),
    ],
)
def test_lint_basic(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        args + [Arg(Option.FLAKE8, "--extend-ignore=E721")],
        expected_output,
        from_empty=False,
    )

def test_lint_file_in_folder() -> None:
    dir_path = get_tests_path(str(Path() / "file"))
    apply_and_lint(
        dir_path,
        [Arg(Option.PYLINT, "--enable=missing-module-docstring")],
        [lazy_problem().set_source(Linter.PYLINT), lazy_problem().set_source(Linter.PYLINT)],
        from_empty=True,
    )


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        (
            "z202817-zkouska.py",
            [Arg(Option.SET_GROUPS, "enhancement")],
            [
                lazy_problem().set_code("R6609").set_line(10),
                lazy_problem().set_code("R6303").set_line(42),
                lazy_problem().set_code("R6208").set_line(76),
                lazy_problem().set_code("R6205").set_line(82),
                lazy_problem().set_code("R6307").set_line(171),
                lazy_problem().set_code("W0107").set_line(196),
            ],
        ),
        (
            "z202817-zkouska.py",
            [Arg(Option.SET_GROUPS, "python-specific")],
            [
                lazy_problem().set_code("R6609").set_line(10),
                lazy_problem().set_code("R6303").set_line(42),
                # lazy_problem().set_code("R6308").set_line(80), TODO differentiate?
                lazy_problem().set_code("R6205").set_line(82),
                lazy_problem().set_code("R6307").set_line(171),
                lazy_problem().set_code("C0123").set_line(172),
                lazy_problem().set_code("W0107").set_line(196),
            ],
        ),
        (
            "z202817-zkouska.py",
            [Arg(Option.PYLINT, "--disable=R,W"), Arg(Option.SET_GROUPS, "python-specific")],
            [
                lazy_problem().set_code("C0123").set_line(172),
            ],
        ),
        (
            "z202817-zkouska.py",
            [Arg(Option.PYLINT, "--disable=R,W,C"), Arg(Option.SET_GROUPS, "")],
            [],
        ),
        (
            "hw34406.py",
            [Arg(Option.PYLINT, "--disable=R,W,C"), Arg(Option.SET_GROUPS, "complexity")],
            [
                lazy_problem().set_line(240).set_code("R0913"),
                lazy_problem().set_line(266).set_code("R0911"),
                lazy_problem().set_line(266).set_code("R0912"),
                lazy_problem().set_line(343).set_code("R0912"),
                lazy_problem().set_line(343).set_code("R0915"),
                lazy_problem().set_line(387).set_code("R1702"),
                lazy_problem().set_line(387).set_code("R1702"),
                lazy_problem().set_line(387).set_code("R1702"),
            ],
        ),
    ],
)
def test_option_sets(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        args + [Arg(Option.FLAKE8, "--extend-ignore=E721")],
        expected_output,
        from_empty=False,
    )


class TestIB111Week:
    @pytest.mark.parametrize(
        "lines,args,expected_output",
        [
            (
                [
                    "def swap(a, b):",
                    "    tmp = a",
                    "    a = b",
                    "    b = tmp",
                ],
                [Arg(Option.SET_GROUPS, "python-specific")],
                [
                    lazy_problem().set_code("R1712").set_line(2),
                ],
            ),
            (
                [
                    "def swap(a, b):",
                    "    tmp = a",
                    "    a = b",
                    "    b = tmp",
                ],
                [
                    Arg(Option.SET_GROUPS, "python-specific"),
                    Arg(Option.SET_GROUPS, "ib111-week-02"),
                ],
                [],
            ),
            (
                [
                    "def swap(a, b):",
                    "    tmp = a",
                    "    a = b",
                    "    b = tmp",
                ],
                [
                    Arg(Option.SET_GROUPS, "python-specific,ib111-week-03"),
                ],
                [
                    lazy_problem().set_code("R1712").set_line(2),
                ],
            ),
            (
                [
                    "def is_one_or_two(n):",
                    "    if n == 1 or n == 2:",
                    "        return True",
                    "    return False",
                ],
                [
                    Arg(Option.SET_GROUPS, "python-specific,ib111-week-06"),
                    Arg(Option.PYLINT, "--disable=R6201"),
                ],
                [],
            ),
            (
                [
                    "def is_one_or_two(n):",
                    "    if n == 1 or n == 2:",
                    "        return True",
                    "    return False",
                ],
                [
                    Arg(Option.SET_GROUPS, "python-specific,ib111-week-07"),
                    Arg(Option.PYLINT, "--disable=R6201"),
                ],
                [
                    lazy_problem().set_code("R1714").set_line(2),
                ],
            ),
        ],
    )
    def test_ib111_week_custom(
        self, lines: List[str], args: List[Arg], expected_output: List[Problem]
    ) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.CONFIG_FILE, "tests/ib111-weeks.toml")] + args,
            expected_output,
            from_empty=False,
        )


@pytest.mark.parametrize(
    "lines,expected_output",
    [
        (["class A:", "    pass"], []),
        (["class _:", "    pass"], []),
        (["class a:", "    pass"], [lazy_problem().set_line(1)]),
        (["class class_name:", "    pass"], [lazy_problem().set_line(1)]),
        (["class CLASS_NAME:", "    pass"], [lazy_problem().set_line(1)]),
        (["def fun():", "    a = 5"], []),
        (["def fun():", "    local_variable = 5"], []),
        (["def fun():", "    _ = 5"], []),
        (["def fun():", "    A = 5"], [lazy_problem().set_line(2)]),
        (["def fun():", "    LOCAL_VARIABLE = 5"], [lazy_problem().set_line(2)]),
        (["def fun():", "    localVariable = 5"], [lazy_problem().set_line(2)]),
        (["def f():", "    pass"], []),
        (["def just_fun():", "    pass"], []),
        (["def _():", "    pass"], []),
        (["def F():", "    pass"], [lazy_problem().set_line(1)]),
        (["def JUST_FUN():", "    pass"], [lazy_problem().set_line(1)]),
        (["def JustFun():", "    pass"], [lazy_problem().set_line(1)]),
        (["A = 5"], []),
        (["GLOBAL_VARIABLE = 1"], []),
        (["TypeAlias = int"], []),
        (["_ = int"], []),
        (["a = 1"], [lazy_problem().set_line(1)]),
        (["global_variable = 1"], [lazy_problem().set_line(1)]),
    ],
)
def test_invalid_name_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(lines, [Arg(Option.PYLINT, "--enable=invalid-name")], expected_output)


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        (
            "014186-p2_nested.py",
            [Arg(Option.SET_GROUPS, "python-specific")],
            [
                lazy_problem().set_code("C0103").set_line(20),
                lazy_problem().set_code("C0103").set_line(21),
                lazy_problem().set_code("C0103").set_line(27),
                lazy_problem().set_code("C0103").set_line(28),
                lazy_problem().set_code("C0103").set_line(31),
                lazy_problem().set_code("C0103").set_line(34),
                lazy_problem().set_code("W0622").set_line(48),
            ],
        ),
        (
            "custom_pep_assign.py",
            [Arg(Option.SET_GROUPS, "python-specific")],
            [
                lazy_problem().set_code("C0103").set_line(1),
            ],
        ),
    ],
)
def test_invalid_name(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output, from_empty=False)


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        (
            "014180-p5_fibsum.py",
            [Arg(Option.ALLOWED_ONECHAR_NAMES, "")],
            [
                lazy_problem().set_code("C0104").set_line(6),
                lazy_problem().set_code("C0104").set_line(14),
            ],
        ),
        (
            "014180-p5_fibsum.py",
            [Arg(Option.ALLOWED_ONECHAR_NAMES, "n")],
            [lazy_problem().set_code("C0104").set_line(14)],
        ),
        (
            "014180-p5_fibsum.py",
            [Arg(Option.ALLOWED_ONECHAR_NAMES, "i")],
            [lazy_problem().set_code("C0104").set_line(6)],
        ),
        ("014180-p5_fibsum.py", [Arg(Option.ALLOWED_ONECHAR_NAMES, "in")], []),
    ],
)
def test_allowed_onechar_names(
    filename: str, args: List[Arg], expected_output: List[Problem]
) -> None:
    apply_and_lint(
        filename, [Arg(Option.PYLINT, "--enable=disallowed-name")] + args, expected_output
    )


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        (
            "105119-p5_template.py",
            [Arg(Option.PYLINT, "--enable=R1714")],
            [
                lazy_problem().set_code("R1714").set_line(22)
                # .set_text("Consider merging these comparisons with 'in' by using 'i not in '[]''"),
                .set_text(
                    "Consider merging these comparisons with 'in' by using 'i not in ('[', ']')'. "
                    "Use a set instead if elements are hashable."
                ),
                lazy_problem()
                .set_code("R1714")
                .set_line(35)
                .set_text(
                    "Consider merging these comparisons with 'in' by using 'i in (1, 2, 3)'. "
                    "Use a set instead if elements are hashable."
                ),
            ],
        ),
    ],
)
def test_consider_using_in(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output)


class TestPyTA:
    @pytest.mark.parametrize(
        "lines,expected_output",
        [
            (["from typing import Tuple", "Point = Tuple[int, int]"], []),
            (["if 0 == 1:", "    print('noooooooo')"], [lazy_problem().set_line(1)]),
        ],
    )
    def test_forbidden_toplevel_code_custom(
        self, lines: List[str], expected_output: List[Problem]
    ) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--enable=forbidden-top-level-code")],
            [p.set_code("E9992") for p in expected_output],
        )


class TestNoGlobals:
    @pytest.mark.parametrize("lines,expected_output", [
        ([
            "x = 0",
            "def foo():",
            "    global x",
            "    x = 1"
        ], [
            lazy_problem() .set_line(1)
            .set_text( "Do not use global variables; you use x, modifying it for example at line 4.")
        ]),
        ([
            "x = 0",
            "def foo():",
            "    nonlocal x",
            "    x = 1"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 4.")
        ]),
        ([
            "x = 0",
            "def foo():",
            "    x = 1",
            "    def bar():",
            "        nonlocal x",
            "        x = 2",
            ], [],
        ),
        ([
            "x = []",
            "def foo():",
            "    x.append(1)"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 3.")
        ]),
        ([
            "x = None",
            "def foo():",
            "    x.y = 1"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 3.")
        ]),
        ([
            "if __name__ == '__main__':",
            "    test = 1",
            "    test = 2",
        ], []
        )
    ])
    def test_no_globals_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--enable=no-global-variables")],
            [p.set_code("R6401") for p in expected_output],
        )

    @pytest.mark.parametrize("filename,expected_output", [
        ("00af22d431-p1_config.py", []),
        ("054141-p1_database.py", []),
        ("054103-p1_database.py", []),
        ("58c81369b3.py", [lazy_problem().set_line(6)]),
        ("cf_1.py", []),
        ("cf_96.py", [lazy_problem().set_line(72)]),
        ("f1dadb36fe-hw3.py", []),
        ("hw14699.py", []),
        ("014466-fibfibsum.py", [
            lazy_problem().set_line(11)
            .set_text("Do not use global variables; you use fib, modifying it for example at line 18.")
        ]),
        ("021752-cards.py", [
            lazy_problem().set_line(7)
            .set_text("Do not use global variables; you use first_num, modifying it for example at line 27."),
            lazy_problem().set_line(8)
            .set_text("Do not use global variables; you use digits, modifying it for example at line 20."),
        ]),
        ("088952-p2_extremes.py", []),
        ("cf_119.py", []),
        ("cf_18.py", [lazy_problem().set_line(13)]),
        ("s10265-d_mancala.py", []),
    ])
    def test_no_globals_files(self, filename: str, expected_output: List[Problem]) -> None:
        apply_and_lint(
            filename, [Arg(Option.PYLINT, "--enable=no-global-variables")], expected_output
        )


class TestLongCode:
    @pytest.mark.parametrize(
        "filename,expected_output",
        [
            (
                "hw48505.py",
                [
                    lazy_problem().set_line(80),
                    lazy_problem().set_line(196),
                ],
            ),
        ],
    )
    def test_long_function_files(self, filename: str, expected_output: List[Problem]) -> None:
        apply_and_lint(filename, [Arg(Option.PYLINT, "--enable=long-function")], expected_output)

    @pytest.mark.parametrize(
        "filename,expected_output",
        [
            ("014422-next.py", [lazy_problem().set_line(19)]),
            ("023240-cards.py", []),
        ],
    )
    def test_use_early_return(self, filename: str, expected_output: List[Problem]) -> None:
        apply_and_lint(filename, [Arg(Option.PYLINT, "--enable=use-early-return")], expected_output)


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        (
            "s2b8861_warehouse.py",
            [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "flake8")],
            [
                lazy_problem().set_line(2).set_text("Forbidden magic comment 'noqa'"),
                lazy_problem().set_line(3).set_text("Forbidden magic comment 'noqa'"),
            ],
        ),
        (
            "s2b8861_warehouse.py",
            [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "all")],
            [
                lazy_problem().set_line(2).set_text("Forbidden magic comment 'noqa'"),
                lazy_problem().set_line(3).set_text("Forbidden magic comment 'noqa'"),
            ],
        ),
        ("s2b8861_warehouse.py", [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "pylint")], []),
        (
            "014422-next-early-return-disabled.py",
            [Arg(Option.PYLINT, "--enable=use-early-return")],
            [],
        ),
        (
            "014422-next-early-return-disabled.py",
            [
                Arg(Option.PYLINT, "--enable=use-early-return"),
                Arg(Option.IGNORE_INFILE_CONFIG_FOR, "pylint"),
            ],
            [lazy_problem().set_line(19).set_text("Forbidden magic comment 'pylint: disable'")],
        ),
        (
            "014422-next-early-return-disabled.py",
            [
                Arg(Option.PYLINT, "--enable=use-early-return"),
                Arg(Option.IGNORE_INFILE_CONFIG_FOR, "flake8"),
            ],
            [],
        ),
        (
            "014422-next-early-return-disabled.py",
            [
                Arg(Option.PYLINT, "--enable=use-early-return"),
                Arg(Option.IGNORE_INFILE_CONFIG_FOR, "all"),
            ],
            [lazy_problem().set_line(19).set_text("Forbidden magic comment 'pylint: disable'")],
        ),
        (
            str(Path("file") / "somewhere"),
            [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "all")],
            []
        )
    ],
)
def test_ignore_infile_config(
    filename: str, args: List[Arg], expected_output: List[Problem]
) -> None:
    apply_and_lint(filename, args, expected_output)


@pytest.mark.parametrize(
    "filename,args,expected_output",
    [
        (
            "umime_count_a.py",
            [
                Arg(Option.SET_GROUPS, "python-specific,enhancement"),
                Arg(Option.ALLOWED_ONECHAR_NAMES, ""),
            ],
            [
                lazy_problem()
                .set_code("C0104")
                .set_line(2)
                .set_text(
                    'Disallowed single-character variable name "a", choose a more descriptive name'
                ),
                lazy_problem().set_code("E225").set_line(2),
                lazy_problem()
                .set_code("R6307")
                .set_line(3)
                .set_text('Iterate directly: "for var in text" (with appropriate name for "var")'),
                lazy_problem().set_code("E225").set_line(4).set_column(19),
                lazy_problem().set_code("E225").set_line(4).set_column(35),
                lazy_problem()
                .set_code("R6609")
                .set_line(5)
                .set_text("Use augmented assignment: 'a += 1'"),
            ],
        ),
        (
            "umime_count_a_extended.py",
            [
                Arg(Option.SET_GROUPS, "python-specific,enhancement"),
                Arg(Option.ALLOWED_ONECHAR_NAMES, ""),
            ],
            [
                lazy_problem()
                .set_code("R6201")
                .set_line(2)
                .set_text("The if statement can be replaced with 'return ch == 'a' or ch == 'A''"),
                lazy_problem().set_code("R1714").set_line(2)
                # .set_text("Consider merging these comparisons with 'in' by using 'ch in 'aA''"),
                .set_text(
                    "Consider merging these comparisons with 'in' by using 'ch in ('a', 'A')'. "
                    "Use a set instead if elements are hashable."
                ),
                lazy_problem()
                .set_code("C0104")
                .set_line(9)
                .set_text(
                    'Disallowed single-character variable name "a", choose a more descriptive name'
                ),
                lazy_problem()
                .set_code("R6307")
                .set_line(10)
                .set_text('Iterate directly: "for var in text" (with appropriate name for "var")'),
                lazy_problem()
                .set_code("R6609")
                .set_line(12)
                .set_text("Use augmented assignment: 'a += 1'"),
            ],
        ),
    ],
)
def test_umime_count_a(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output, from_empty=False)


@pytest.mark.parametrize(
    "lines,expected_output",
    [
        (["def fun():", "    return 42", "    pass"], [lazy_problem().set_code("W0101")]),
    ],
)
def test_overrides_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(lines, [Arg(Option.NO_FLAKE8, None)], expected_output, from_empty=False)


@pytest.mark.parametrize("filename", ["hs3013-8548.py"])
def test_lints_without_error(filename: str) -> None:
    just_lint(filename, [], from_empty=False)


@pytest.mark.filterwarnings(
    "ignore:The 'default' argument to fields is deprecated. Use 'dump_default' instead."
)
def test_problem_can_be_dumped_to_json() -> None:
    problem = Problem(
        source=Linter.FLAKE8,
        enabled_by="foo",
        path="path",
        line=5,
        column=1,
        code="E303",
        text="too many blank lines (3)",
        end_line=None,
        end_column=None,
    )
    out = problem.to_json(indent=2, sort_keys=True)  # type: ignore
    assert (
        out
        == """{
  "code": "E303",
  "column": 1,
  "enabled_by": "foo",
  "end_column": null,
  "end_line": null,
  "line": 5,
  "path": "path",
  "source": "flake8",
  "symbol": null,
  "text": "too many blank lines (3)"
}"""
    )

    out = Problem.schema().dumps([problem], indent=2, many=True, sort_keys=True)  # type: ignore
    assert (
        out
        == """[
  {
    "code": "E303",
    "column": 1,
    "enabled_by": "foo",
    "end_column": null,
    "end_line": null,
    "line": 5,
    "path": "path",
    "source": "flake8",
    "symbol": null,
    "text": "too many blank lines (3)"
  }
]"""
    )



# From: https://github.com/PyCQA/pylint/blob/v2.16.2/doc/exts/pylint_messages.py

# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/PyCQA/pylint/blob/main/CONTRIBUTORS.txt

"""Script used to generate the messages files."""

import os
from collections import defaultdict
from inspect import getmodule
from itertools import chain, groupby
from pathlib import Path
from typing import DefaultDict, Dict, List, NamedTuple, Optional, Tuple

from sphinx.application import Sphinx

from pylint.checkers import initialize as initialize_checkers
from pylint.constants import MSG_TYPES
from pylint.extensions import initialize as initialize_extensions
from pylint.lint import PyLinter
from pylint.message import MessageDefinition
from pylint.utils import get_rst_title

PYLINT_BASE_PATH = Path(__file__).resolve().parent.parent.parent
"""Base path to the project folder."""

PYLINT_MESSAGES_PATH = PYLINT_BASE_PATH / "doc/user_guide/messages"
"""Path to the messages documentation folder."""

PYLINT_MESSAGES_DATA_PATH = PYLINT_BASE_PATH / "doc" / "data" / "messages"
"""Path to the folder with data for the messages documentation."""

MSG_TYPES_DOC = {k: v if v != "info" else "information" for k, v in MSG_TYPES.items()}


class MessageData(NamedTuple):
    checker: str
    id: str
    name: str
    definition: MessageDefinition
    good_code: str
    bad_code: str
    details: str
    related_links: str
    checker_module_name: str
    checker_module_path: str
    shared: bool = False
    default_enabled: bool = True


MessagesDict = Dict[str, List[MessageData]]
OldMessagesDict = Dict[str, DefaultDict[Tuple[str, str], List[Tuple[str, str]]]]
"""DefaultDict is indexed by tuples of (old name symbol, old name id) and values are
tuples of (new name symbol, new name category).
"""


def _register_all_checkers_and_extensions(linter: PyLinter) -> None:
    """Registers all checkers and extensions found in the default folders."""
    initialize_checkers(linter)
    initialize_extensions(linter)


def _get_message_data(data_path: Path) -> Tuple[str, str, str, str]:
    """Get the message data from the specified path."""
    good_py_path = data_path / "good.py"
    bad_py_path = data_path / "bad.py"
    details_rst_path = data_path / "details.rst"
    related_rst_path = data_path / "related.rst"
    if not data_path.exists():
        _create_placeholders(data_path, details_rst_path, good_py_path)
    good_code = _get_titled_rst(title="Correct code", text=_get_python_code_as_rst(good_py_path))
    bad_code = _get_titled_rst(title="Problematic code", text=_get_python_code_as_rst(bad_py_path))
    details = _get_titled_rst(title="Additional details", text=_get_rst_as_str(details_rst_path))
    related = _get_titled_rst(title="Related links", text=_get_rst_as_str(related_rst_path))
    _check_placeholders(bad_code, details, good_py_path, related)
    return good_code, bad_code, details, related


def _check_placeholders(bad_code: str, details: str, good_py_path: Path, related: str) -> None:
    if bad_code or related:
        placeholder_details = "help us make the doc better" in details
        with open(good_py_path, encoding="utf8") as f:
            placeholder_good = "placeholder" in f.read()
        assert_msg = (
            f"Please remove placeholders in '{good_py_path.parent}' "
            f"as you started completing the documentation"
        )
        assert not placeholder_good and not placeholder_details, assert_msg


def _get_titled_rst(title: str, text: str) -> str:
    """Return rst code with a title if there is anything in the section."""
    return f"**{title}:**\n\n{text}" if text else ""


def _get_rst_as_str(rst_path: Path) -> str:
    """Return the content of an 'rst' file or an empty string if the file does not
    exist.
    """
    if not rst_path.exists():
        return ""
    with open(rst_path, encoding="utf-8") as f:
        return f.read()


def _get_python_code_as_rst(code_path: Path) -> str:
    """Return the 'rst' representation of a python file or an empty string if the file
    does not exist.
    """
    if not code_path.exists():
        return ""
    return f"""\
.. literalinclude:: /{code_path.relative_to(Path.cwd())}
   :language: python
"""


def _create_placeholders(data_path: Path, details_rst_path: Path, good_py_path: Path) -> None:
    data_path.mkdir(parents=True)
    with open(good_py_path, "w", encoding="utf-8") as file:
        file.write(
            """\
# This is a placeholder for correct code for this message.
"""
        )
    with open(details_rst_path, "w", encoding="utf-8") as file:
        file.write(
            """\
You can help us make the doc better `by contributing <https://github.com/PyCQA/pylint/issues/5953>`_ !
"""
        )


def _get_all_messages(
    linter: PyLinter,
) -> Tuple[MessagesDict, OldMessagesDict]:
    """Get all messages registered to a linter and return a dictionary indexed by
    message type.

    Also return a dictionary of old message and the new messages they can be mapped to.
    """
    messages_dict: MessagesDict = {
        "fatal": [],
        "error": [],
        "warning": [],
        "convention": [],
        "refactor": [],
        "information": [],
    }
    old_messages: OldMessagesDict = {
        "fatal": defaultdict(list),
        "error": defaultdict(list),
        "warning": defaultdict(list),
        "convention": defaultdict(list),
        "refactor": defaultdict(list),
        "information": defaultdict(list),
    }
    checker_message_mapping = chain.from_iterable(
        ((checker, msg) for msg in checker.messages) for checker in linter.get_checkers()
    )

    for checker, message in checker_message_mapping:
        good_code, bad_code, details, related = _get_message_data(_get_message_data_path(message))

        checker_module = getmodule(checker)

        assert (
            checker_module and checker_module.__file__
        ), f"Cannot find module for checker {checker}"

        message_data = MessageData(
            message.checker_name,
            message.msgid,
            message.symbol,
            message,
            good_code,
            bad_code,
            details,
            related,
            checker_module.__name__,
            checker_module.__file__,
            message.shared,
            message.default_enabled,
        )
        msg_type = MSG_TYPES_DOC[message.msgid[0]]
        messages_dict[msg_type].append(message_data)
        if message.old_names:
            for old_name in message.old_names:
                category = MSG_TYPES_DOC[old_name[0][0]]
                # We check if the message is already in old_messages so
                # we don't duplicate shared messages.
                if (message.symbol, msg_type) not in old_messages[category][
                    (old_name[1], old_name[0])
                ]:
                    old_messages[category][(old_name[1], old_name[0])].append(
                        (message.symbol, msg_type)
                    )

    return messages_dict, old_messages


def _get_message_data_path(message: MessageDefinition) -> Path:
    return PYLINT_MESSAGES_DATA_PATH / message.symbol[0] / message.symbol


def _message_needs_update(message_data: MessageData, category: str) -> bool:
    """Do we need to regenerate this message .rst ?"""
    message_path = _get_message_path(category, message_data)
    if not message_path.exists():
        return True
    message_path_stats = message_path.stat().st_mtime
    checker_path_stats = Path(message_data.checker_module_path).stat().st_mtime
    return checker_path_stats > message_path_stats


def _get_category_directory(category: str) -> Path:
    return PYLINT_MESSAGES_PATH / category


def _get_message_path(category: str, message: MessageData) -> Path:
    category_dir = _get_category_directory(category)
    return category_dir / f"{message.name}.rst"


def _write_message_page(messages_dict: MessagesDict) -> None:
    """Create or overwrite the file for each message."""
    for category, messages in messages_dict.items():
        category_dir = _get_category_directory(category)
        if not category_dir.exists():
            category_dir.mkdir(parents=True, exist_ok=True)
        for message in messages:
            if message.shared:
                continue
            if not _message_needs_update(message, category):
                continue
            _write_single_message_page(category_dir, message)
        for _, shared_messages in groupby(
            sorted((message for message in messages if message.shared), key=lambda m: m.id),
            key=lambda m: m.id,
        ):
            shared_messages_list = list(shared_messages)
            if len(shared_messages_list) > 1:
                _write_single_shared_message_page(category_dir, shared_messages_list)
            else:
                _write_single_message_page(category_dir, shared_messages_list[0])


def _generate_single_message_body(message: MessageData) -> str:
    body = f""".. _{message.name}:

{get_rst_title(f"{message.name} / {message.id}", "=")}
**Message emitted:**

{message.definition.msg}

**Description:**

*{message.definition.description}*
"""
    if not message.default_enabled:
        body += f"""
.. caution::
  This message is disabled by default. To enable it, add ``{message.name}`` to the ``enable`` option.

"""

    body += f"""
{message.bad_code}
{message.good_code}
{message.details}
{message.related_links}
"""
    if message.checker_module_name.startswith("pylint.extensions."):
        body += f"""
.. note::
  This message is emitted by the optional :ref:`'{message.checker}'<{message.checker_module_name}>`
   checker which requires the ``{message.checker_module_name}`` plugin to be loaded.

"""
    return body


def _generate_checker_url(message: MessageData) -> str:
    checker_module_rel_path = os.path.relpath(message.checker_module_path, PYLINT_BASE_PATH)
    return f"https://github.com/PyCQA/pylint/blob/main/{checker_module_rel_path}"


def _write_single_shared_message_page(category_dir: Path, messages: List[MessageData]) -> None:
    message = messages[0]
    with open(category_dir / f"{message.name}.rst", "w", encoding="utf-8") as stream:
        stream.write(_generate_single_message_body(message))
        checker_urls = ", ".join(
            [f"`{message.checker} <{_generate_checker_url(message)}>`__" for message in messages]
        )
        stream.write(f"Created by the {checker_urls} checkers.")


def _write_single_message_page(category_dir: Path, message: MessageData) -> None:
    with open(category_dir / f"{message.name}.rst", "w", encoding="utf-8") as stream:
        stream.write(_generate_single_message_body(message))
        checker_url = _generate_checker_url(message)
        stream.write(f"Created by the `{message.checker} <{checker_url}>`__ checker.")


def _write_messages_list_page(
    messages_dict: MessagesDict, old_messages_dict: OldMessagesDict
) -> None:
    """Create or overwrite the page with the list of all messages."""
    messages_file = os.path.join(PYLINT_MESSAGES_PATH, "messages_overview.rst")
    with open(messages_file, "w", encoding="utf-8") as stream:
        # Write header of file
        title = "Messages overview"
        stream.write(
            f"""
.. _messages-overview:

{"#" * len(title)}
{get_rst_title(title, "#")}

.. This file is auto-generated. Make any changes to the associated
.. docs extension in 'doc/exts/pylint_messages.py'.

Pylint can emit the following messages:

"""
        )
        # Iterate over tuple to keep same order
        for category in (
            "fatal",
            "error",
            "warning",
            "convention",
            "refactor",
            "information",
        ):
            # We need to remove all duplicated shared messages
            messages = sorted(
                {msg.id: msg for msg in messages_dict[category]}.values(),
                key=lambda item: item.name,
            )
            old_messages = sorted(old_messages_dict[category], key=lambda item: item[0])
            messages_string = "".join(f"   {category}/{message.name}\n" for message in messages)
            old_messages_string = "".join(
                f"   {category}/{old_message[0]}\n" for old_message in old_messages
            )
            # Write list per category. We need the '-category' suffix in the reference
            # because 'fatal' is also a message's symbol
            stream.write(
                f"""
.. _{category.lower()}-category:

{get_rst_title(category.capitalize(), "*")}
All messages in the {category} category:

.. toctree::
   :maxdepth: 2
   :titlesonly:

{messages_string}
All renamed messages in the {category} category:

.. toctree::
   :maxdepth: 1
   :titlesonly:

{old_messages_string}"""
            )


def _write_redirect_pages(old_messages: OldMessagesDict) -> None:
    """Create redirect pages for old-messages."""
    for category, old_names in old_messages.items():
        category_dir = PYLINT_MESSAGES_PATH / category
        if not os.path.exists(category_dir):
            os.makedirs(category_dir)
        for old_name, new_names in old_names.items():
            _write_redirect_old_page(category_dir, old_name, new_names)


def _write_redirect_old_page(
    category_dir: Path,
    old_name: Tuple[str, str],
    new_names: List[Tuple[str, str]],
) -> None:
    old_name_file = os.path.join(category_dir, f"{old_name[0]}.rst")
    new_names_string = "".join(f"   ../{new_name[1]}/{new_name[0]}.rst\n" for new_name in new_names)
    content = f""".. _{old_name[0]}:

{get_rst_title("/".join(old_name), "=")}
"{old_name[0]} has been renamed. The new message can be found at:

.. toctree::
   :maxdepth: 2
   :titlesonly:

{new_names_string}
"""
    with open(old_name_file, "w", encoding="utf-8") as stream:
        stream.write(content)


# pylint: disable-next=unused-argument
def build_messages_pages(app: Optional[Sphinx]) -> None:
    """Overwrite messages files by printing the documentation to a stream.

    Documentation is written in ReST format.
    """
    # Create linter, register all checkers and extensions and get all messages
    linter = PyLinter()
    _register_all_checkers_and_extensions(linter)
    messages, old_messages = _get_all_messages(linter)

    # Write message and category pages
    _write_message_page(messages)
    _write_messages_list_page(messages, old_messages)

    # Write redirect pages
    _write_redirect_pages(old_messages)


def setup(app: Sphinx) -> None:
    """Connects the extension to the Sphinx process."""
    # Register callback at the builder-inited Sphinx event
    # See https://www.sphinx-doc.org/en/master/extdev/appapi.html
    app.connect("builder-inited", build_messages_pages)


if __name__ == "__main__":
    pass
    # Uncomment to allow running this script by your local python interpreter
    # build_messages_pages(None)



from typing import Dict, Any, Optional, List
from dataclasses import dataclass, field
from loguru import logger

from edulint.linters import Linter
from edulint.config.utils import config_file_val_to_str
from edulint.option_parses import OPTION_SETS_LABEL


@dataclass
class OptionSet:
    to: Dict[Linter, List[str]] = field(default_factory=dict)
    description: str = ""


OptionSets = Dict[str, OptionSet]


def parse_option_sets(raw_option_sets: Any) -> Optional[OptionSets]:
    if not isinstance(raw_option_sets, dict):
        logger.warning(
            "option sets are not a dictionary but a value {val} of type {type}",
            val=raw_option_sets,
            type=type(raw_option_sets),
        )
        return None

    option_sets = {}
    for name, to in raw_option_sets.items():
        if not isinstance(to, dict):
            logger.warning(
                "option set named {name} is not a dictionary but a value {val} of type {type}",
                name=name,
                val=to,
                type=type(to),
            )
            continue

        option_set = OptionSet()
        for linter_str, translated in to.items():
            linter = Linter.safe_from_name(linter_str)
            if linter is None:
                logger.warning(
                    "invalid value {linter} where one of {linters} is expected",
                    linter=linter_str,
                    linters=", ".join(linter.to_name() for linter in Linter),
                )
                continue

            to_process = (
                [translated] if not isinstance(translated, dict) else list(translated.items())
            )
            result = []
            for processed in to_process:
                str_val = config_file_val_to_str(OPTION_SETS_LABEL, processed)
                if str_val:
                    result.append(str_val)

            if len(result) > 0:
                option_set.to[linter] = result

        if len(option_set.to) > 0:
            option_sets[name] = option_set

    return option_sets
